#include "ControllerParser.h"

#include <iostream>
#include <fstream>
#include <map>

#include <boost/optional.hpp>
#include <boost/phoenix/operator.hpp>
#include <boost/phoenix/object/construct.hpp>
#include <boost/algorithm/string.hpp>

BOOST_FUSION_ADAPT_STRUCT(
    controller::transition,
    (std::string, pre)
    (std::string, input)
    (std::string, post)
)

namespace controller {

template <typename Iterator>
ControllerParser::transition_parser<Iterator>::transition_parser() : transition_parser::base_type(transition_rule, "transition") {
    quoted_string %= '"' > *~qi::char_('"') > '"';
    transition_rule %= qi::eps > quoted_string > qi::lit("->") > quoted_string > qi::lit("->") > quoted_string;

    transition_rule.name("transition");

    qi::on_error<qi::fail>
    (
        transition_rule
      , std::cerr
            << phx::val("Error: Expecting ")
            << qi::_4                                        // what failed?
            << phx::val(" here: \"")
            << phx::construct<std::string>(qi::_3, qi::_2)   // iterators to error-pos, end
            << phx::val("\"")
            << std::endl
    );
}

ControllerParser::ControllerParser() :
    parser()
{ }

ControllerParser::~ControllerParser() { }

int ControllerParser::parse_line(unsigned int line_number, std::string& line, transition& transition) const {
    iterator_type begin = line.begin();
    iterator_type end   = line.end();
    bool r = qi::phrase_parse(begin, end, parser, qi::space, transition);
    if (r && begin != end) {
        std::cerr << "Error: Trailing unparsed: " << std::string(begin, end) << std::endl;
        r = false;
    }
    if (!r) {
        std::cerr << "Error: Parsing failed on line " << line_number << ": " << line << std::endl;
        return -1;
    }
    return 1;
}

boost::optional<Controller> ControllerParser::parse(const std::string& filename) const {
    std::ifstream in(filename);
    if (!in.is_open()) {
        std::cerr << "Error: Could not open input file: " << filename << std::endl;
        return boost::none;
    }
    in.unsetf(std::ios::skipws);

    std::string line;
    std::vector<transition> transitions{};

    // read transitions
    unsigned int line_number = 0;
    // discard first line
    std::getline(in, line);
    while (std::getline(in, line)) {
        line_number += 1;
        transition transition;
        int r = parse_line(line_number, line, transition);
        if (r < 0) {
            return boost::none;
        }
        else if (r > 0) {
            transitions.push_back(transition);
        }
    }
    in.close();

    return transitions;
}

}
