#include "MPGArena.h"

#include <iostream>
#include <iterator>
#include <map>

namespace mpg {

Node::Node(id_t id, Player owner, boost::optional< std::string > name):
    id(id),
    owner(owner),
    name(name)
{ }

Node::~Node() {
    remove_edges();
    unlink();
}

void Node::unlink() {
    NodeHook::unlink();
}

unsigned int Node::remove_edges() {
    unsigned int count = 0;
    succ_edges.clear_and_dispose( [&count] (auto x) { count++; delete x; } );
    pred_edges.clear_and_dispose( [&count] (auto x) { count++; delete x; } );
    return count;
}

Edge::Edge(Node* src, Node* dst, weight_t weight) :
    src(src),
    dst(dst),
    weight(weight)
{
    src->succ_edges.push_back(*this);
    dst->pred_edges.push_back(*this);
}
Edge::~Edge() {
    unlink();
}

void Edge::unlink() {
    EdgeSrcHook::unlink();
    EdgeDstHook::unlink();
}


MPGArena::MPGArena() :
    nodes(),
    n_nodes(0),
    n_edges(0),
    min_weight(WEIGHT_MAX),
    max_weight(WEIGHT_MIN)
{
}

MPGArena::~MPGArena() {
    nodes.clear_and_dispose( [] (auto x) { delete x; } );
}

Node* MPGArena::add_node(id_t id, Player owner, boost::optional< std::string > name) {
    Node* node = new Node(id, owner, name);
    nodes.push_back(*node);
    n_nodes++;
    return node;
}

Edge* MPGArena::add_edge(Node* src, Node* dst, weight_t weight) {
    Edge* edge = new Edge(src, dst, weight);
    n_edges++;
    if (weight > max_weight) {
        max_weight = weight;
    }
    if (weight < min_weight) {
        min_weight = weight;
    }
    return edge;
}

void MPGArena::remove_node(Node* node) {
    n_edges -= node->remove_edges();
    delete node;
    n_nodes--;
}

void MPGArena::remove_edge(Edge* edge) {
    delete edge;
    n_edges--;
}


void MPGArena::print() const {
    for (const auto &n : nodes) {
        std::cout << n.id << " ";
        if (n.owner == Player::MAX) {
            std::cout << "MAX";
        }
        else {
            std::cout << "MIN";
        }
        if (n.name) {
            std::cout << " \"" << *n.name << "\"";
        }
        std::cout << " [";
        auto pred_it = n.pred_edges.begin();
        auto pred_end = n.pred_edges.end();
        while (pred_it != pred_end) {
            std::cout << pred_it->src->id;
            if (pred_it->weight > 0) {
                std::cout << "+";
            }
            else if (pred_it->weight == 0) {
                std::cout << "±";
            }
            std::cout << pred_it->weight;
            ++pred_it;
            if (pred_it != pred_end) {
                std::cout << ",";
            }
        }
        std::cout << "] [";
        auto succ_it = n.succ_edges.begin();
        auto succ_end = n.succ_edges.end();
        while (succ_it != succ_end) {
            std::cout << succ_it->dst->id << ":";
            if (succ_it->weight > 0) {
                std::cout << "+";
            }
            else if (succ_it->weight == 0) {
                std::cout << "±";
            }
            std::cout << succ_it->weight;
            ++succ_it;
            if (succ_it != succ_end) {
                std::cout << ",";
            }
        }
        std::cout << "]" << std::endl;
    }
}

void MPGArena::print_basic_info() const {
    std::cout << "Number of nodes: " << n_nodes << std::endl;
    std::cout << "Number of edges: " << n_edges << std::endl;
}

void MPGArena::print_file(std::ostream& out) const {
    for (const auto &n : nodes) {
        out << n.id << " ";
        if (n.owner == Player::MAX) {
            out << "MAX ";
        }
        else {
            out << "MIN ";
        }
        auto succ_it = n.succ_edges.begin();
        auto succ_end = n.succ_edges.end();
        while (succ_it != succ_end) {
            out << succ_it->dst->id << ":" << succ_it->weight;
            ++succ_it;
            if (succ_it != succ_end) {
                out << ",";
            }
        }
        if (n.name) {
            out << " \"" << *n.name << "\"";
        }
        out << ";" << std::endl;
    }
}

}
