#include "MPGArenaParser.h"

#include <iostream>
#include <fstream>
#include <map>

#include <boost/phoenix/operator.hpp>
#include <boost/phoenix/object/construct.hpp>
#include <boost/algorithm/string.hpp>

BOOST_FUSION_ADAPT_STRUCT(
    mpg::MPGArenaParser::edge,
    (mpg::id_t, dst)
    (mpg::weight_t, weight)
)

BOOST_FUSION_ADAPT_STRUCT(
    mpg::MPGArenaParser::node,
    (mpg::id_t, id)
    (mpg::Player, owner)
    (std::vector<mpg::MPGArenaParser::edge>, successors)
    (boost::optional< std::string >, name)
)

namespace mpg {

template <typename Iterator>
MPGArenaParser::node_parser<Iterator>::node_parser() : node_parser::base_type(node_rule, "node") {
    quoted_string %= '"' > *~qi::char_('"') > '"';
    id_rule %= qi::uint_;
    owner_rule %= qi::lit("MAX")[qi::_val=Player::MAX] | qi::lit("MIN")[qi::_val=Player::MIN];
    weight_rule %= qi::long_long;
    edge_rule %= id_rule > ':' > weight_rule;
    node_rule %= qi::eps > id_rule > owner_rule > edge_rule % ',' > (-quoted_string) > ';';

    id_rule.name("id");
    owner_rule.name("owner");
    weight_rule.name("weight");
    edge_rule.name("edge");
    node_rule.name("node");

    qi::on_error<qi::fail>
    (
        node_rule
      , std::cerr
            << phx::val("Error: Expecting ")
            << qi::_4                                        // what failed?
            << phx::val(" here: \"")
            << phx::construct<std::string>(qi::_3, qi::_2)   // iterators to error-pos, end
            << phx::val("\"")
            << std::endl
    );
}

template <typename Iterator>
MPGArenaParser::comment_parser<Iterator>::comment_parser() : comment_parser::base_type(comment_rule, "comment") {
    comment_rule = '#' > *qi::char_;

    comment_rule.name("comment");

    qi::on_error<qi::fail>(comment_rule, qi::_1);
}

MPGArenaParser::MPGArenaParser() :
    parser(),
    comm_parser()
{ }

MPGArenaParser::~MPGArenaParser() { }

int MPGArenaParser::parse_line(unsigned int line_number, std::string& line, node& node) const {
    boost::algorithm::trim_left(line);
    iterator_type begin = line.begin();
    iterator_type end   = line.end();
    qi::phrase_parse(begin, end, comm_parser, qi::space);
    if (begin == end) {
        return 0;
    }
    else {
        bool r = qi::phrase_parse(begin, end, parser, qi::space, node);
        qi::phrase_parse(begin, end, comm_parser, qi::space);
        if (r && begin != end) {
            std::cerr << "Error: Trailing unparsed: " << std::string(begin, end) << std::endl;
            r = false;
        }
        if (!r) {
            std::cerr << "Error: Parsing failed on line " << line_number << ": " << line << std::endl;
            return -1;
        }
        return 1;
    }
}

boost::optional<MPGArena*> MPGArenaParser::parse(const std::string& filename) const {
    std::ifstream in(filename);
    if (!in.is_open()) {
        std::cerr << "Error: Could not open input file: " << filename << std::endl;
        return boost::none;
    }
    in.unsetf(std::ios::skipws);

    std::string line;
    std::vector<node> nodes{};

    // read nodes
    unsigned int line_number = 0;
    while (std::getline(in, line)) {
        line_number += 1;
        node node;
        int r = parse_line(line_number, line, node);
        if (r < 0) {
            return boost::none;
        }
        else if (r > 0) {
            nodes.push_back(node);
        }
    }
    in.close();

    MPGArena* arena = new MPGArena();
    std::map< unsigned int, Node* > node_map;
    for (auto node : nodes) {
        Node* arena_node = arena->add_node(node.id, node.owner, node.name);
        node_map[node.id]  = arena_node;
    }
    for (auto node : nodes) {
        Node* src = node_map[node.id];
        for (auto edge : node.successors) {
            Node* dst = node_map[edge.dst];
            arena->add_edge(src, dst, edge.weight);
        }
    }
    return arena;
}

}
