#include "OptionParser.h"

#include <iostream>

namespace mpgsolver {

OptionParser::OptionParser() :
    help_options_desc_("Options"),
    hidden_options_desc_(),
    positional_options_desc_(),
    options_desc_()
{
    help_options_desc_.add_options()
        ("help,h", "display this help text and exit")
        ("input-file-format,i", po::value<std::string>(), "input file format: mpg, pg, hoa or bin")
        ("output-file,o", po::value<std::string>(), "output file")
        ("cycle-check,c", "explicitly check for non-positive cycles in the game graph")
        ("scc-partition", "compute the scc partition of the game arena")
        ("no-compact-colors", "do not compact the colors by removing unused colors")
        ("no-auxilliary-nodes", "do not add auxilliary nodes to the arena (may lead to wrong results)")
        ("sort-by-owner", "sort nodes by owner (within scc with scc partitions)")
        ("device,d", po::value<std::string>(), "select the device (by index or name)")
        ("work-group-size,w", po::value<size_t>(), "set the work group size (0=auto)")
        ("print-arena,p", "print the arena")
        ("print-arena-info", "print basic arena information")
        ("device-info", "print device information")
        ("profiling", "measure and print kernel profiling information")
        ("timing,t", "measure and print timing information")
        ("no-solving", "do not solve the game")
        ("zero-mean-partition", "only solve the zero mean partition problem")
        ("check-realizability", "check realizability")
    ;
    hidden_options_desc_.add_options()
        ("input-file", po::value<std::vector<std::string>>(), "input file")
    ;
    po::options_description options;
    options_desc_.add(help_options_desc_).add(hidden_options_desc_);
    positional_options_desc_.add("input-file", -1);
}

OptionParser::~OptionParser() { }

void OptionParser::print_options() const {
    std::cout << help_options_desc_;
}

Options OptionParser::parse_options(const int argc, const char* argv[]) const {
    po::variables_map vm;

    po::store(po::command_line_parser(argc, argv).
            options(options_desc_).
            positional(positional_options_desc_).
            run(), vm);
    po::notify(vm);

    Options options;

    options.help = vm.count("help") > 0;
    options.check_for_cycles = vm.count("cycle-check") > 0;
    options.scc_partition = vm.count("scc-partition") > 0;
    options.compact_colors = vm.count("no-compact-colors") == 0;
    options.auxilliary_nodes = vm.count("no-auxilliary-nodes") == 0;
    options.sort_by_owner = vm.count("sort-by-owner") > 0;
    options.solve_game = vm.count("no-solving") == 0;
    options.zero_mean_partition = vm.count("zero-mean-partition") > 0;
    options.check_realizability = vm.count("check-realizability") > 0;
    options.print_arena = vm.count("print-arena") > 0;
    options.print_arena_info = vm.count("print-arena-info") > 0;
    options.print_device_info = vm.count("device-info") > 0;
    options.profiling = vm.count("profiling") > 0;
    options.timing = vm.count("timing") > 0;

    if (vm.count("device")) {
        options.device = vm["device"].as<std::string>();
    } else {
        options.device = "";
    }

    if (vm.count("work-group-size")) {
        options.work_group_size = vm["work-group-size"].as<size_t>();
    } else {
        options.work_group_size = 0;
    }

    if (vm.count("input-file")) {
        options.input_files = vm["input-file"].as<std::vector<std::string>>();
    } else {
        options.input_files = {};
    }

    if (vm.count("input-file-format")) {
        std::string input_file_format = vm["input-file-format"].as<std::string>();
        boost::algorithm::to_lower(input_file_format);
        if (input_file_format == "mpg") {
            options.input_file_format = InputFileFormat::MPG;
        }
        else if (input_file_format == "pg") {
            options.input_file_format = InputFileFormat::PG;
        }
        else if (input_file_format == "hoa") {
            options.input_file_format = InputFileFormat::HOA;
        }
        else if (input_file_format == "bin") {
            options.input_file_format = InputFileFormat::BIN;
        }
        else {
            throw std::invalid_argument("Invalid input file format selected: " + vm["input-file-format"].as<std::string>());
        }
    } else {
        options.input_file_format = InputFileFormat::MPG;
    }

    if (vm.count("output-file")) {
        options.output_file = vm["output-file"].as<std::string>();
    } else {
        options.output_file = "";
    }

    return options;
}

}
