#include <iostream>
#include <fstream>
#include <vector>

#include <boost/optional.hpp>

#include "OptionParser.h"
#include "util/Timer.h"
#include "parser/Arena.h"
#include "parser/MPGArenaParser.h"
#include "parser/NodePGArenaParser.h"
#include "parser/HOPGParser.h"
#include "parser/BINParser.h"
#include "ocl/OCLDriver.h"
#include "ocl/arena/OCLArena.h"
#include "ocl/preprocessor/OCLPreprocessor.h"
#include "ocl/solver/OCLSolver.h"
#include "ocl/solver/OCLMPGSolver.h"
#include "ocl/solver/OCLPGSolver.h"

int main(const int argc, const char* argv[]) {
    try {

        // Parse options
        mpgsolver::OptionParser option_parser;
        mpgsolver::Options options = option_parser.parse_options(argc, argv);

        if (options.help) {
            std::cout << "Usage: " << argv[0] << " [OPTIONS] FILES" << std::endl;
            std::cout << "Solve the mean-payoff game in FILES on the GPU." << std::endl;
            std::cout << std::endl;
            option_parser.print_options();
            return EXIT_SUCCESS;
        }

        // Start main part
        Timer total_timer(options.timing);
        Timer timer(options.timing);
        total_timer.start("invoking mpgsolver");

        timer.start("selecting device");
        ocl::OCLDriver* driver = new ocl::OCLDriver(options.device, options.profiling, options.work_group_size);
        if (options.print_device_info) {
            driver->printDeviceInformation();
        }
        timer.stop();

        ocl::OCLArena* arena = nullptr;
        ocl::OCLPreprocessor* preprocessor = nullptr;
        ocl::OCLSolver* solver = nullptr;

        timer.start("parsing input file");
        if (options.input_file_format == mpgsolver::InputFileFormat::BIN) {
            if (options.input_files.size() < 2) {
               throw std::invalid_argument("Two input files are need for the binary format, for nodes and edges.");
            }
            parser::BINParser parser;
            parser::BINParseResult* parsed_arena = parser.parse(options.input_files[0], options.input_files[1]);
            arena = new ocl::OCLArena(driver, *parsed_arena);
            delete parsed_arena;
        }
        else {
            parser::Arena* parsed_arena = nullptr;
            if (options.input_files.empty()) {
                throw std::invalid_argument("No input file was given.");
            }
            if (options.input_file_format == mpgsolver::InputFileFormat::MPG) {
                parser::MPGArenaParser parser;
                parsed_arena = parser.parse(options.input_files[0]);
            }
            else if (options.input_file_format == mpgsolver::InputFileFormat::PG) {
                parser::NodePGArenaParser parser;
                parsed_arena = parser.parse(options.input_files[0]);
            }
            else if (options.input_file_format == mpgsolver::InputFileFormat::HOA) {
                parser::HOPGParser parser;
                parsed_arena = parser.parse(options.input_files[0]);
            }
            else {
                throw std::invalid_argument("Unsupported input format");
            }
            arena = new ocl::OCLArena(driver, *parsed_arena);
            delete parsed_arena;
        }
        timer.stop();

        timer.start("preprocessing arena");
        arena->initArena(options.scc_partition, options.compact_colors, options.auxilliary_nodes, options.sort_by_owner);
        timer.stop();

        timer.start("initializing solver and driver");
        if (options.solve_game) {
            if (options.input_file_format == mpgsolver::InputFileFormat::MPG) {
                solver = new ocl::OCLMPGSolver(driver, arena, options.check_for_cycles, options.zero_mean_partition);
            }
            else if (
                    options.input_file_format == mpgsolver::InputFileFormat::PG ||
                    options.input_file_format == mpgsolver::InputFileFormat::HOA ||
                    options.input_file_format == mpgsolver::InputFileFormat::BIN
            ) {
                solver = new ocl::OCLPGSolver(driver, arena, options.check_realizability);
            }
        }
        driver->init();
        timer.stop();

        if (options.print_arena_info) {
            timer.start("writing arena info");
            arena->print_basic_info();
            timer.stop();
        }
        if (options.solve_game) {
            timer.start("solving game");
            solver->execute();
            timer.stop();
        }

        std::streambuf* buf;
        std::ofstream of;
        if (options.output_file.empty()) {
            buf = std::cout.rdbuf();
            options.output_file = "stdout";
        }
        else {
            of.open(options.output_file, std::ios::out);
            if (!of) {
                throw std::invalid_argument("Could not open output file: " + options.output_file);
            }
            buf = of.rdbuf();
        }
        std::ostream out(buf);
        if (options.print_arena) {
            timer.start("writing arena");
            arena->print(out);
            timer.stop();
        }
        if (options.solve_game) {
            timer.start("writing solution");
            solver->print_solution(out);
            timer.stop();
        }
        if (options.profiling) {
            driver->printProfilingInformation();
        }
        of.close();

        delete solver;
        delete preprocessor;
        delete driver;
        delete arena;

        total_timer.stop();
    }
    catch (const boost::compute::opencl_error e) {
        std::cerr << "Error: OpenCL error: " << e.error_string() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::exception& e) {
        std::cerr << "Error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::string& e) {
        std::cerr << "Error: " << e << std::endl;
        return EXIT_FAILURE;
    }
    catch (...) {
        std::cerr << "Error: Unknown error." << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
