#include "OCLDriver.h"

#include <iostream>
#include <fstream>
#include <sstream>

namespace ocl {

OCLDriver::OCLDriver(const std::string& device_selection, const bool profiling, const size_t work_group_size) :
    profiling(profiling),
    work_size(0),
    work_group_size(0)
{
    if ((work_group_size & (work_group_size - 1)) != 0) {
        throw std::invalid_argument("work group size needs to be 0 or a power of 2");
    }

    selectDevice(device_selection);

    context = compute::context(device);
    cl_command_queue_properties command_queue_properties = 0;
    if (profiling) {
        command_queue_properties |= compute::command_queue::enable_profiling;
    }
    queue = compute::command_queue(context, device, command_queue_properties);

    addConstant("WORK_GROUP_SIZE", work_group_size);

    addSourceFile("ocl/OCLDefinitions.h");

    // create return value buffer
    return_value_buffer = new compute::buffer(context, sizeof(cl_bool));
}

OCLDriver::~OCLDriver() {
    delete return_value_buffer;
}

void OCLDriver::init() {
    if (work_size == 0) {
        throw std::logic_error("Work size not set!");
    }
    if (work_group_size != 0) {
        // ensure that work size is a multiple of work group size
        work_size = (work_size + work_group_size - 1) & ~(work_group_size - 1);
    }

    std::string source{};
    for (const auto& source_file : source_files) {
        loadKernelSourceFile(source, source_file);
    }
    compute::program program = compute::program::create_with_source(source, context);
    std::stringstream build_options_stream;
    build_options_stream << "-Werror -DCL_COMPILE -cl-kernel-arg-info";
#ifdef NDEBUG
    // propagate NDEBUG flag
    build_options_stream << " -DNDEBUG";
#endif
    // set definitions and constants as flags
    for (const auto definition : definitions) {
        build_options_stream << " -D" << definition;
    }
    for (const auto constant : constants) {
        build_options_stream << " -D" << constant.first << "=" << constant.second;
    }
    try {
        program.build(build_options_stream.str());
    }
    catch(compute::opencl_error &e) {
        throw std::runtime_error("Could not build OpenCL program:\n" + program.build_log());
    }

    // load kernels
    for (auto& kernel : kernels) {
        kernel->loadKernel(program);
    }
}

void OCLDriver::loadKernelSourceFile(std::string& source, const std::string& filename) const {
    std::string path(std::string(CL_SOURCE_DIR) + "/" + filename);
    std::ifstream input_stream(path, std::ios::in);
    if (!input_stream) {
        throw std::invalid_argument("Unable to open kernel file: " + path);
    }
    input_stream.seekg(0, std::ios::end);
	size_t current_size = source.size();
    size_t source_code_size = input_stream.tellg();
	source.resize(current_size + source_code_size);
    input_stream.seekg(0, std::ios::beg);
	input_stream.read(&source[current_size], source_code_size);
    input_stream.close();
}


void OCLDriver::addSourceFile(const std::string& source_file) {
    source_files.push_back(source_file);
}

OCLKernel* OCLDriver::addKernel(const std::string& name) {
    OCLKernel* kernel = new OCLKernel(name);
    kernels.push_back(kernel);
    return kernel;
}

void OCLDriver::addConstant(const std::string& name, unsigned int value) {
    constants.push_back({name, value});
}

void OCLDriver::addDefinition(const std::string& name) {
    definitions.push_back(name);
}

void OCLDriver::setWorkSize(size_t size) {
    work_size = size;
}

void OCLDriver::executeVoidKernel(OCLKernel* kernel) {
    compute::event event = queue.enqueue_1d_range_kernel(kernel->kernel, 0, work_size, work_group_size);
    event.wait();
    if (profiling) {
        kernel->count++;
        kernel->duration += event.duration<boost::chrono::nanoseconds>();
    }
}

bool OCLDriver::executeBoolKernel(OCLKernel* kernel) {
    cl_bool return_value = false;
    queue.enqueue_write_buffer(*return_value_buffer, 0, sizeof(cl_bool), &return_value);
    executeVoidKernel(kernel);
    queue.enqueue_read_buffer(*return_value_buffer, 0, sizeof(cl_bool), &return_value);
    return return_value;
}

void OCLDriver::printProfilingInformation() const {
    if (!profiling) {
        throw std::logic_error("Profiling not enabled.");
    }
    typedef boost::chrono::duration<double> seconds;

    std::cout << std::setw(115) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
    std::cout << "Profiling information:" << std::endl;
    std::cout << std::setw(115) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
    std::cout << std::left << std::setw(40) << "Kernel" << std::right
              << std::setw(15) << "Call count"
              << std::setw(20) << "Duration (s)"
              << std::setw(20) << "Duration (%)"
              << std::setw(20) << "Average (s)" << std::endl;
    unsigned int total_count = 0;
    boost::chrono::nanoseconds total_duration(0);
    for (const auto& kernel : kernels) {
        total_count += kernel->count;
        total_duration += kernel->duration;
    }
    for (const auto& kernel : kernels) {
        std::cout << std::left << std::setw(40) << kernel->name << std::right
              << std::setw(15) << kernel->count
              << std::setw(20) << std::fixed << std::setprecision(9) << boost::chrono::duration_cast<seconds>(kernel->duration).count()
              << std::setw(20) << std::fixed << std::setprecision(2) << (100.0 * kernel->duration.count()/total_duration.count())
              << std::setw(20) << std::fixed << std::setprecision(9);
        if (kernel->count != 0) {
            std::cout << (boost::chrono::duration_cast<seconds>(kernel->duration).count() / kernel->count);
        }
        else {
            std::cout << "N/A";
        }
        std::cout << std::endl;
    }
    std::cout << std::setw(115) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
    std::cout << std::left << std::setw(40) << "Total" << std::right
              << std::setw(15) << total_count
              << std::setw(20) << std::fixed << std::setprecision(9) << boost::chrono::duration_cast<seconds>(total_duration).count()
              << std::setw(20) << std::fixed << std::setprecision(2) << 100.0
              << std::setw(20) << std::fixed << std::setprecision(9)
              << (boost::chrono::duration_cast<seconds>(total_duration).count() / total_count) << std::endl;
    std::cout << std::setw(115) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
}

void OCLDriver::printDeviceInformation() const {
        std::cout << std::setw(80) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
        std::cout << "Device information" << std::endl;
        std::cout << std::setw(80) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
        std::cout << "Vendor                : " << device.vendor() << std::endl;
        std::cout << "Name                  : " << device.name() << std::endl;
        std::cout << "Profile               : " << device.profile() << std::endl;
        std::cout << "Version               : " << device.version() << std::endl;
        std::cout << "Driver version        : " << device.driver_version() << std::endl;
        std::cout << "Type                  : ";
        if ((device.type() & CL_DEVICE_TYPE_CPU) != 0) {
            std::cout << "CPU";
        }
        else if ((device.type() & CL_DEVICE_TYPE_GPU) != 0) {
            std::cout << "GPU";
        }
        else if ((device.type() & CL_DEVICE_TYPE_ACCELERATOR) != 0) {
            std::cout << "Accelerator";
        }
        if ((device.type() & CL_DEVICE_TYPE_DEFAULT) != 0) {
            std::cout << " [Default]";
        }
        std::cout << std::endl;
        std::cout << "Adress bits           : " << device.address_bits() << std::endl;
        std::cout << "Global memory size    : " << device.global_memory_size() << std::endl;
        std::cout << "Local memory size     : " << device.local_memory_size() << std::endl;
        std::cout << "Max mem alloc size    : " << device.get_info<CL_DEVICE_MAX_MEM_ALLOC_SIZE>() << std::endl;
        std::cout << "Clock frequency       : " << device.clock_frequency() << std::endl;
        std::cout << "Compute units         : " << device.compute_units() << std::endl;
        std::cout << "Max work group size   : " << device.get_info<CL_DEVICE_MAX_WORK_GROUP_SIZE>() << std::endl;
        std::cout << std::setw(80) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
        std::cout << "Supported extensions" << std::endl;
        std::cout << std::setw(80) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
        for (const auto& extension : device.extensions()) {
            if (!extension.empty()) {
                std::cout << extension << std::endl;
            }
        }
        std::cout << std::setw(80) << std::setfill('-') << "-" << std::setfill(' ') << std::endl;
}

void OCLDriver::selectDevice(const std::string& device_selection) {
    typedef const boost::iterator_range<std::string::const_iterator> StringRange;

    std::cout << "Checking available devices …" << std::endl;
    std::vector<boost::compute::device> devices = boost::compute::system::devices();
    if (devices.size() == 0) {
        throw std::runtime_error("No OpenCL devices found. Check OpenCL installation!");
    }
    size_t device_id;
    if (device_selection.empty()) {
        device_id = 0;
    }
    else {
        try {
            device_id = std::stoi(device_selection);
        }
        catch (...) {
            device_id = std::string::npos;
        }
    }
    std::cout << "Available devices:" << std::endl;
    size_t i = 0;
    for (auto it = devices.cbegin(); it != devices.cend(); ++i, ++it) {
        std::stringstream device_name_stream;
        device_name_stream <<
            it->vendor() << " " <<
            it->name() <<
            " [" << it->profile() << "] " <<
            it->version() <<
            " (" << it->driver_version() << ")";
        //std::string device_name =
        std::string device_name = device_name_stream.str();
        std::cout << "  [" << i << "] "<< device_name;
        if (
                i == device_id || (device_id == std::string::npos &&
                boost::ifind_first(StringRange(device_name.begin(), device_name.end()),
                                   StringRange(device_selection.begin(), device_selection.end())))
        ) {
            std::cout << " [Selected]";
            device = *it;
            device_id = i;
        }
        std::cout << std::endl;
    }
    if (device_id == std::string::npos || device_id >= devices.size()) {
        throw std::invalid_argument("Invalid device selected: " + device_selection);
    }
}


}
