#pragma once

#if defined(__APPLE__)
#include <OpenCL/cl.cpp>
#else
#include <CL/cl.hpp>
#endif

#include <map>

#include <boost/chrono/include.hpp>

#include "OCLDefinitions.h"

namespace ocl {

class OCLKernel {
friend class OCLDriver;
private:
    const std::string name;
    compute::kernel kernel;
    unsigned int count;
    boost::chrono::nanoseconds duration;
    std::map< std::string, size_t> argument_map;

    OCLKernel(const std::string& name) :
        name(name),
        count(0),
        duration(0)
    { }

    void loadKernel(compute::program program) {
        kernel = program.create_kernel(name);
        size_t arity = kernel.arity();
        for (size_t i = 0; i < arity; ++i) {
            const std::string arg_name = kernel.get_arg_info<std::string>(i, CL_KERNEL_ARG_NAME);
            argument_map.insert({arg_name, i});
        }
    }

public:
    template<typename T>
    void setArg(const std::string& arg_name, const T& value) {
        kernel.set_arg(argument_map.at(arg_name), value);
    }
};

class OCLDriver {
private:
    const bool profiling;

    size_t work_size;
    const size_t work_group_size;

    std::vector<std::string> source_files;
    std::vector<OCLKernel*> kernels;
    std::vector< std::pair< std::string, unsigned int > > constants;
    std::vector< std::string > definitions;

    void loadKernelSourceFile(std::string& source, const std::string& filename) const;
    void selectDevice(const std::string& device_selection);

public:
    compute::device device;
    compute::context context;
    compute::command_queue queue;

    compute::buffer* return_value_buffer;

    OCLDriver(const std::string& device_selection = "", bool profiling = false, size_t work_group_size = 0);
    ~OCLDriver();

    void addSourceFile(const std::string& source_file);
    OCLKernel* addKernel(const std::string& name);
    void addConstant(const std::string& name, unsigned int value);
    void addDefinition(const std::string& name);

    void setWorkSize(size_t size);

    void executeVoidKernel(OCLKernel* kernel);
    bool executeBoolKernel(OCLKernel* kernel);

    void init();
    void clean();

    void printProfilingInformation() const;
    void printDeviceInformation() const;
};

}

