#include "OCLArena.h"

#include <iostream>
#include <iomanip>
#include <sstream>
#include <cmath>
#include <limits>
#include <algorithm>
#include <cstdint>

namespace ocl {

OCLArena::OCLArena(
        OCLDriver* driver,
        const parser::Arena& arena
) :
    OCLModule("OCL Arena"),
    driver(driver),
    n_nodes(arena.n_nodes),
    n_edges(arena.n_edges),
    n_colors(0),
    n_sccs(0),
    n_nodes_auxilliary_fwd(0),
    n_nodes_auxilliary_bwd(0),
    has_transpose_graph(false),
    has_n_succs(true),
    has_sccs(false),
    has_edge_weights(arena.has_edge_weights),
    has_node_colors(arena.has_node_colors),
    has_edge_colors(arena.has_edge_colors),
    parity_type(arena.parity_type)
{
    if (n_nodes == 0) {
        throw std::logic_error("No nodes!");
    }
    if (n_nodes >= std::numeric_limits<id_t>::max()) {
        throw std::logic_error("Too many nodes:" + std::to_string(n_nodes));
    }
    if (n_edges >= std::numeric_limits<id_t>::max()) {
        throw std::logic_error("Too many edges: " + std::to_string(n_edges));
    }
    if (has_node_colors && has_edge_colors) {
        throw std::logic_error("both node and edge colors not supported");
    }

    ids = new id_t[n_nodes];
    bits = new bits_t[n_nodes];
    succs_begin = new id_t[n_nodes + 1];
    n_succs = new id_t[n_nodes];
    succs = new id_t[n_edges];

    if (has_edge_weights) {
        weights = new weight_t[n_edges];
    }
    if (has_node_colors) {
        colors = new color_t[n_nodes];
    }
    if (has_edge_colors) {
        colors = new color_t[n_edges];
    }

    initial_node = -1;

    size_t i = 0;
    size_t j = 0;
    for (const auto& node : arena.nodes) {
        ids[i] = node.id;
        if (node.owner == parser::PLAYER_0) {
            bits[i] = PLAYER_P0;
        }
        else {
            bits[i] = PLAYER_P1;
        }
        succs_begin[i] = j;
        n_succs[i] = node.successors.size();
        if (has_node_colors) {
            colors[i] = node.color;
            if (node.color >= n_colors && node.color != parser::COLOR_UNDEFINED) {
                n_colors = node.color + 1;
            }
        }

        for (const auto& edge : node.successors) {
            id_t successor = arena.id_map.at(edge.dst);
            succs[j] = successor;
            if (has_edge_weights) {
                weights[j] = edge.weight;
            }
            if (has_edge_colors) {
                colors[j] = edge.color;
                if (edge.color >= n_colors && edge.color != parser::COLOR_UNDEFINED) {
                    n_colors = edge.color + 1;
                }
            }
            j++;
        }
        i++;
    }
    succs_begin[n_nodes] = n_edges;

    if (parity_type == parser::ParityType::MAX) {
        neutral_color = 0;
    }
    else {
        neutral_color = n_colors - 1;
    }
    if (arena.has_undefined_colors) {
        size_t n = 0;
        if (has_node_colors) {
            n = n_nodes;
        }
        else if (has_edge_colors) {
            n = n_edges;
        }
        for (size_t i = 0; i < n; ++i) {
            if (colors[i] == parser::COLOR_UNDEFINED) {
                colors[i] = neutral_color;
            }
        }
    }
}

OCLArena::OCLArena(
        OCLDriver* driver,
        const parser::BINParseResult& arena
) :
    OCLModule("OCL Arena"),
    driver(driver),
    n_nodes(arena.n_nodes),
    n_edges(arena.n_edges),
    n_colors(arena.n_colors),
    n_sccs(0),
    n_nodes_auxilliary_fwd(0),
    n_nodes_auxilliary_bwd(0),
    has_transpose_graph(false),
    has_n_succs(false),
    has_sccs(false),
    has_edge_weights(false),
    has_node_colors(false),
    has_edge_colors(true),
    parity_type(parser::ParityType::MIN),
    neutral_color(arena.n_colors - 1)
{
    ids = arena.ids;
    bits = arena.bits;
    succs_begin = arena.succs_begin;
    succs = arena.succs;
    colors = arena.colors;

    initial_node = arena.initial_node;
}

void OCLArena::initArena(const bool scc_partition, const bool compact_colors, const bool auxilliary_nodes, const bool sort_by_owner) {
    if (scc_partition) {
        computeSCCs();
    }
    else {
        addSingleSCC();
    }
    if (n_sccs > 1 || sort_by_owner) {
        sortNodes(sort_by_owner);
    }
    if (compact_colors && (has_node_colors || has_edge_colors)) {
        compactColors();
    }
    if (auxilliary_nodes) {
        computeAuxilliaryNodes();
    }

    setDriverSettings();
}

OCLArena::~OCLArena() {
    delete[] ids;
    delete[] bits;
    delete[] succs_begin;
    delete[] succs;

    if (has_n_succs) {
        delete[] n_succs;
    }
    if (has_edge_weights) {
        delete[] weights;
    }
    if (has_node_colors || has_edge_colors) {
        delete[] colors;
    }
    if (has_transpose_graph) {
        delete[] preds_begin;
        delete[] n_preds;
        delete[] preds;
    }
    delete[] sccs_begin;
    delete[] scc_index;
}

void OCLArena::clean() {
    delete bits_buffer;
}

void OCLArena::setDriverSettings() {
    driver->addSourceFile("ocl/arena/OCLArenaDefinitions.h");

    driver->addConstant("N_NODES", n_nodes);
    driver->addConstant("N_EDGES", n_edges);
    driver->addConstant("N_COLORS", n_colors);
    if (has_edge_weights) {
        driver->addDefinition("EDGE_WEIGHTS");
    }
    if (has_node_colors) {
        driver->addDefinition("NODE_COLORS");
    }
    if (has_edge_colors) {
        driver->addDefinition("EDGE_COLORS");
    }
    if (parity_type == parser::ParityType::MAX) {
        driver->addDefinition("MAX_PARITY");
    }
    else {
        driver->addDefinition("MIN_PARITY");
    }

    driver->setWorkSize(n_nodes);
}

void OCLArena::computeAuxilliaryNodes() {
    for (size_t k = 0; k < n_sccs; ++k) {
        size_t i_begin = sccs_begin[k];
        size_t i_end = sccs_begin[k+1];
        for (size_t i = i_begin; i != i_end; ++i) {
            for (size_t j = succs_begin[i]; j != succs_begin[i+1]; ++j) {
                // TODO compare fwd and bwd edges for each scc
                id_t dst = succs[j];
                if ((bits[dst] & MASK_OWNER) == (bits[i] & MASK_OWNER)) {
                    if (dst <= i) {
                    //if (dst <= i && dst >= i_begin) { TODO also adapt in kernels, set edge flag?
                        n_nodes_auxilliary_bwd++;
                        SET_BWD_AUX_NODE(bits[i]);
                    }
                    //if (dst >= i && dst < i_end) {
                    if (dst >= i) {
                        n_nodes_auxilliary_fwd++;
                        SET_FWD_AUX_NODE(bits[i]);
                    }
                }
            }
        }
    }
    // only use auxilliary nodes in one direction
    if (n_nodes_auxilliary_bwd > n_nodes_auxilliary_fwd) {
        n_nodes_auxilliary_bwd = 0;
    }
    else {
        n_nodes_auxilliary_fwd = 0;
    }
    driver->addConstant("N_NODES_AUXILLIARY_FWD", n_nodes_auxilliary_fwd);
    driver->addConstant("N_NODES_AUXILLIARY_BWD", n_nodes_auxilliary_bwd);
}

void OCLArena::computeTransposeGraph() {
    preds_begin = new id_t[n_nodes + 1];
    n_preds = new id_t[n_nodes];
    preds = new id_t[n_edges];

    std::fill( n_preds, n_preds + n_nodes, 0 );
    for (size_t j = 0; j < n_edges; ++j) {
        ++n_preds[succs[j]];
    }
    id_t cur_preds_begin = 0;
    for (size_t i = 0; i < n_nodes; ++i) {
        id_t cur_n_preds = n_preds[i];
        preds_begin[i] = cur_preds_begin;
        cur_preds_begin += cur_n_preds;
        n_preds[i] = 0;
    }
    preds_begin[n_nodes] = n_edges;

    for (size_t i = 0; i < n_nodes; ++i) {
        for (size_t j = succs_begin[i]; j < succs_begin[i+1]; ++j) {
            id_t successor = succs[j];
            id_t k = preds_begin[successor] + n_preds[successor];
            preds[k] = i;
            n_preds[successor]++;
        }
    }
    has_transpose_graph = true;
}

void OCLArena::computeNumberOfSuccessors() {
    n_succs = new id_t[n_nodes];

    for (size_t i = 0; i < n_nodes; ++i) {
        n_succs[i] = succs_begin[i+1] - succs_begin[i];
        if (n_succs[i] == 0) { // TODO remove warnings
            std::cerr << "Warning: node " << ids[i] << " (";
            if (IS_P0_NODE(bits[i])) {
                std::cerr << "p0";
            }
            else {
                std::cerr << "p1";
            }
            std::cerr << ") has no successors" << std::endl;
        }
    }

    has_n_succs = true;
}

void OCLArena::sortNodes(const bool sort_by_owner) {
    size_t n = n_sccs;
    if (sort_by_owner) {
        n *= 2;
    }

    // compute new permutation
    id_t* count = new id_t[n];
    id_t* node_permutation = new id_t[n_nodes];
    id_t* inverse_permutation = new id_t[n_nodes];

    std::fill( count, count + n, 0 );
    for (size_t i = 0; i < n_nodes; ++i) {
        id_t k = scc_index[i];
        if (sort_by_owner) {
            k += bits[i];
        }
        ++count[k];
    }

    id_t cur_count_begin = 0;
    for (size_t k = 0; k < n; ++k) {
        id_t cur_count = count[k];
        count[k] = cur_count_begin;
        cur_count_begin += cur_count;
    }

    for (size_t i = 0; i < n_nodes; ++i) {
        id_t k = scc_index[i];
        if (sort_by_owner) {
            k += bits[i];
        }
        id_t new_i = count[k];
        node_permutation[i] = new_i;
        inverse_permutation[new_i] = i;
        ++count[k];
    }

    // reorder arrays
    if (initial_node < n_nodes) {
        initial_node = node_permutation[initial_node];
    }

    id_t* new_ids = new id_t[n_nodes];
    bits_t* new_bits = new bits_t[n_nodes];
    id_t* new_scc_index = new id_t[n_nodes];
    id_t* new_succs_begin = new bits_t[n_nodes + 1];

    id_t* new_succs = new bits_t[n_edges];
    id_t* new_n_succs = nullptr;
    if (has_n_succs) {
        new_n_succs = new id_t[n_nodes];
    }

    weight_t* new_weights = nullptr;
    if (has_edge_weights) {
        new_weights = new weight_t[n_edges];
    }
    color_t* new_colors = nullptr;
    if (has_node_colors) {
        new_colors = new color_t[n_nodes];
    }
    else if (has_edge_colors) {
        new_colors = new color_t[n_edges];
    }

    id_t new_j = 0;
    for (size_t new_i = 0; new_i < n_nodes; ++new_i) {
        id_t old_i = inverse_permutation[new_i];
        new_ids[new_i] = ids[old_i];
        new_bits[new_i] = bits[old_i];
        new_scc_index[new_i] = scc_index[old_i];
        new_succs_begin[new_i] = new_j;
        if (has_n_succs) {
            new_n_succs[new_i] = n_succs[old_i];
        }
        for (size_t old_j = succs_begin[old_i]; old_j != succs_begin[old_i+1]; ++old_j) {
            new_succs[new_j] = node_permutation[succs[old_j]];
            if (has_edge_weights) {
                new_weights[new_j] = weights[old_j];
            }
            if (has_edge_colors) {
                new_colors[new_j] = colors[old_j];
            }
            ++new_j;
        }
        if (has_node_colors) {
            new_colors[new_i] = colors[old_i];
        }
    }
    new_succs_begin[n_nodes] = n_edges;

    delete[] node_permutation;
    delete[] inverse_permutation;

    delete[] ids;
    ids = new_ids;

    delete[] bits;
    bits = new_bits;

    delete[] scc_index;
    scc_index = new_scc_index;

    delete[] succs_begin;
    succs_begin = new_succs_begin;

    delete[] succs;
    succs = new_succs;

    if (has_n_succs) {
        delete[] n_succs;
        n_succs = new_n_succs;
    }

    if (has_edge_weights) {
        delete[] weights;
        weights = new_weights;
    }
    if (has_node_colors || has_edge_colors) {
        delete[] colors;
        colors = new_colors;
    }
}

void OCLArena::addSingleSCC() {
    // simply create one over-abstracted SCC with all nodes
    n_sccs = 1;
    sccs_begin = new id_t[2];
    sccs_begin[0] = 0;
    sccs_begin[1] = n_nodes;

    scc_index = new id_t[n_nodes];
    std::fill( scc_index, scc_index + n_nodes, 0 );
}

void OCLArena::computeSCCs() {
    scc_index = new id_t[n_nodes];

    // Tarjan's algorithm for computing SCCs in
    // bottom-up topological order, with own stack
    id_t cur_index = 1;
    id_t* index = new id_t[n_nodes];
    id_t* lowlink = new id_t[n_nodes];
    id_t* stack = new id_t[n_nodes];
    bool* on_stack = new bool[n_nodes];
    id_t* stack_pointer = stack;
    std::fill( index, index + n_nodes, 0 );

    struct stack_frame {
        id_t node;
        id_t edge;
    };
    stack_frame* call_stack = new stack_frame[n_nodes];
    stack_frame* call_stack_pointer = call_stack;

    id_t* tmp_sccs_begin = new id_t[n_nodes];
    id_t cur_scc_begin = 0;
    id_t cur_scc_id = 0;

    id_t i = 0;
    id_t v = 0;
    id_t u;
    id_t w;
    bool new_call = true;
    while (i < n_nodes) {
        if (new_call) {
            index[v] = cur_index;
            lowlink[v] = cur_index;
            ++cur_index;
            *stack_pointer = v;
            ++stack_pointer;
            on_stack[v] = true;
            u = succs_begin[v];
            new_call = false;
        }
        else {
            if (call_stack_pointer != call_stack) {
                --call_stack_pointer;
                stack_frame frame = *call_stack_pointer;
                v = frame.node;
                u = frame.edge;
                w = succs[u];
                lowlink[v] = std::min(lowlink[v], lowlink[w]);
                ++u;
            }
            else {
                while (i < n_nodes && index[i] != 0) {
                    ++i;
                }
                v = i;
                new_call = true;
            }
        }

        while (!new_call && u != succs_begin[v+1]) {
            w = succs[u];
            if (index[w] == 0) {

                stack_frame call_frame = { v, u };
                *call_stack_pointer = call_frame;
                ++call_stack_pointer;

                v = w;

                new_call = true;
            }
            else if (on_stack[w]) {
                lowlink[v] = std::min(lowlink[v], index[w]);
            }
            ++u;
        }

        if (!new_call) {
            if (lowlink[v] == index[v]) {
                id_t w;
                tmp_sccs_begin[cur_scc_id] = cur_scc_begin;
                do {
                    --stack_pointer;
                    w = *stack_pointer;
                    on_stack[w] = false;
                    scc_index[w] = cur_scc_id;
                    ++cur_scc_begin;
                }
                while (v != w);
                ++cur_scc_id;
            }
        }
    }

    delete[] index;
    delete[] lowlink;
    delete[] stack;
    delete[] on_stack;

    delete[] call_stack;

    // copy results to main scc array
    n_sccs = cur_scc_id;
    sccs_begin = new id_t[n_sccs + 1];
    std::copy(tmp_sccs_begin, tmp_sccs_begin + n_sccs, sccs_begin);
    sccs_begin[n_sccs] = n_nodes;

    delete[] tmp_sccs_begin;

    has_sccs = true;
}

void OCLArena::compactColors() {
    id_t* color_count = new id_t[n_colors];
    id_t* color_sorted;
    if (has_node_colors) {
        color_sorted = new id_t[n_nodes];
    }
    else {
        color_sorted = new id_t[n_edges];
    }

    color_t highest_color = 0;

    for (size_t k = 0; k < n_sccs; ++k) {
        size_t i_begin = sccs_begin[k];
        size_t i_end = sccs_begin[k+1];
        if (has_edge_colors) {
            i_begin = succs_begin[i_begin];
            i_end = succs_begin[i_end];
        }

        std::fill( color_count, color_count + n_colors, 0 );
        for (size_t i = i_begin; i != i_end; ++i) {
            ++color_count[colors[i]];
        }

        bool colors_not_compact = false;
        id_t cur_color_begin = i_begin;
        for (size_t c = 0; c < n_colors; ++c) {
            id_t cur_color_count = color_count[c];
            if (cur_color_count == 0) {
                colors_not_compact = true;
            }
            color_count[c] = cur_color_begin;
            cur_color_begin += cur_color_count;
        }

        color_t cur_color;
        if (colors_not_compact) {
            for (size_t i = i_begin; i != i_end; ++i) {
                color_t c = colors[i];
                color_sorted[color_count[c]] = i;
                ++color_count[c];
            }

            cur_color = 0;
            cur_color_begin = i_begin;
            for (size_t c = 0; c < n_colors; ++c) {
                id_t cur_color_end = color_count[c];
                if (cur_color_begin != cur_color_end) {
                    if ((c & 1) != (cur_color & 1)) {
                        ++cur_color;
                    }
                    while (cur_color_begin != cur_color_end) {
                        colors[color_sorted[cur_color_begin]] = cur_color;
                        ++cur_color_begin;
                    }
                }
            }
        }
        else {
            cur_color = n_colors - 1;

        }

        if (cur_color > highest_color) {
            highest_color = cur_color;
        }
    }

    n_colors = highest_color + 1;
    if (parity_type == parser::ParityType::MAX) {
        neutral_color = 0;
    }
    else {
        neutral_color = highest_color;
    }

    delete[] color_sorted;
    delete[] color_count;
}

void OCLArena::init() {
    std::cout << "init arena ocl buffer" << std::endl;
    bits_buffer           = new compute::vector<bits_t>(n_nodes, driver->context);

    compute::copy(bits, bits + n_nodes, bits_buffer->begin(), driver->queue);
}

void OCLArena::print_basic_info() const {
    std::cout << "Number of nodes: " << n_nodes << std::endl;
    std::cout << "Number of edges: " << n_edges << std::endl;
    std::cout << "Number of colors: " << n_colors << std::endl;
    std::cout << "Number of sccs: " << n_sccs << std::endl;
    std::cout << "Auxilliary nodes: " << (n_nodes_auxilliary_fwd + n_nodes_auxilliary_bwd) << std::endl;
}

void OCLArena::print(std::ostream& out) const {
    id_t max_id = n_nodes;
    weight_t max_weight = 0;
    if (has_edge_weights) {
        for (size_t j = 0; j < n_edges; ++j) {
            max_weight = std::max(max_weight, std::abs(weights[j]));
        }
    }
    size_t id_size = 1 + (size_t)log10(max_id);
    size_t weight_size = 0;
    if (has_edge_weights) {
        weight_size = 2 + (size_t)log10(max_weight);
    }
    size_t color_size = 0;
    if (has_node_colors || has_edge_colors) {
        color_size = 1 + (size_t)log10(n_colors);
    }
    id_t cur_scc = 0;
    for (size_t i = 0; i < n_nodes; ++i) {
        if (i >= sccs_begin[cur_scc]) {
            out << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
            ++cur_scc;
        }
        if (i == initial_node) {
            out << ">";
        }
        else {
            out << " ";
        }
        out << std::setw(id_size) << ids[i];
        out << " (" << std::setw(id_size) << scc_index[i] << ")";
        if (IS_P0_NODE(bits[i])) {
            out << " 0";
        }
        else {
            out << " 1";
        }
        if (has_node_colors) {
            out << ":" << std::setw(color_size) << colors[i];
        }
        out << " [";
        for (size_t j = succs_begin[i]; j != succs_begin[i+1]; ++j) {
            out << std::setw(id_size) << ids[succs[j]];
            if (has_edge_weights) {
                out << ":";
                if (weights[j] == 0) {
                    out << std::setw(weight_size) << "0";
                }
                else {
                    out << std::setw(weight_size) << std::showpos << weights[j];
                }
            }
            if (has_edge_colors) {
                out << ":";
                out << std::setw(color_size) << colors[j];
            }
            if (j + 1 != succs_begin[i+1]) {
                out << ",";
            }
        }
        out << "]";
        if (has_transpose_graph) {
            out << " [";
            for (size_t j = preds_begin[i]; j != preds_begin[i] + n_preds[i]; ++j) {
                out << std::setw(id_size) << ids[preds[j]];
                if (j + 1 != preds_begin[i] + n_preds[i]) {
                    out << ",";
                }
            }
            out << "]";
        }
        out << std::endl;
    }
    out << std::setfill('-') << std::setw(20) << "" << std::setfill(' ') << std::endl;
}

}
