#include "OCLPreprocessor.h"

#include <iostream>
#include <iomanip>

namespace ocl {

OCLPreprocessor::OCLPreprocessor(OCLDriver* driver, OCLArena* arena, const bool compute_scc_partition) :
    OCLModule("OCL MPG Preprocessor", 1),
    driver(driver),
    arena(arena),
    compute_scc_partition(compute_scc_partition)
{
    if (compute_scc_partition) {
        if (!arena->has_transpose_graph) {
            arena->computeTransposeGraph();
        }
        if (!arena->has_n_succs) {
            arena->computeNumberOfSuccessors();
        }

        driver->addSourceFile("ocl/preprocessor/OCLPreprocessorDefinitions.h");
        driver->addSourceFile("ocl/preprocessor/OCLPreprocessor.cl");

        kernel_bfs_init = driver->addKernel("bfs_init");
        kernel_bfs_fwd_iteration = driver->addKernel("bfs_fwd_iteration");
        kernel_bfs_bwd_iteration = driver->addKernel("bfs_bwd_iteration");
        kernel_scc_partition_init = driver->addKernel("scc_partition_init");
        kernel_scc_partition_update_values = driver->addKernel("scc_partition_update_values");
        kernel_scc_partition_update_edges = driver->addKernel("scc_partition_update_edges");
        kernel_trivial_sccs = driver->addKernel("trivial_sccs");

        hashtable_size = arena->n_nodes; // TODO use next highest power of two
        driver->addConstant("HASHTABLE_SIZE", hashtable_size);
        scc_ids = new id_t[arena->n_nodes];
    }

    addDependency(arena);
}

OCLPreprocessor::~OCLPreprocessor() {
    if (compute_scc_partition) {
        delete kernel_bfs_init;
        delete kernel_bfs_fwd_iteration;
        delete kernel_bfs_bwd_iteration;
        delete kernel_scc_partition_init;
        delete kernel_scc_partition_update_values;
        delete kernel_scc_partition_update_edges;
        delete kernel_trivial_sccs;

        delete[] scc_ids;
    }
}

void OCLPreprocessor::init() {
    if (compute_scc_partition) {
        succs_begin_buffer   = new compute::vector<id_t>(arena->n_nodes, driver->context);
        preds_begin_buffer   = new compute::vector<id_t>(arena->n_nodes, driver->context);
        n_succs_buffer       = new compute::vector<id_t>(arena->n_nodes, driver->context);
        n_preds_buffer       = new compute::vector<id_t>(arena->n_nodes, driver->context);
        succs_buffer         = new compute::vector<id_t>(arena->n_edges, driver->context);
        preds_buffer         = new compute::vector<id_t>(arena->n_edges, driver->context);

        pivots_buffer        = new compute::vector<id_t>(hashtable_size, driver->context);
        scc_ids_buffer       = new compute::vector<id_t>(arena->n_nodes, driver->context);
        scc_values_buffer    = new compute::vector<scc_value>(arena->n_nodes, driver->context);
        scc_intervals_buffer = new compute::vector<scc_interval>(arena->n_nodes, driver->context);

        compute::copy(arena->succs_begin, arena->succs_begin + arena->n_nodes, succs_begin_buffer->begin(), driver->queue);
        compute::copy(arena->n_succs, arena->n_succs + arena->n_nodes, n_succs_buffer->begin(), driver->queue);
        compute::copy(arena->succs, arena->succs + arena->n_edges, succs_buffer->begin(), driver->queue);
        compute::copy(arena->preds_begin, arena->preds_begin + arena->n_nodes, preds_begin_buffer->begin(), driver->queue);
        compute::copy(arena->n_preds, arena->n_preds + arena->n_nodes, n_preds_buffer->begin(), driver->queue);
        compute::copy(arena->preds, arena->preds + arena->n_edges, preds_buffer->begin(), driver->queue);

        kernel_bfs_init->setArg("bits", *arena->bits_buffer);
        kernel_bfs_init->setArg("scc_values", *scc_values_buffer);
        kernel_bfs_init->setArg("scc_ids", *scc_ids_buffer);
        kernel_bfs_init->setArg("pivots", *pivots_buffer);

        kernel_bfs_fwd_iteration->setArg("change", *driver->return_value_buffer);
        kernel_bfs_fwd_iteration->setArg("bits", *arena->bits_buffer);
        kernel_bfs_fwd_iteration->setArg("succs_begin", *succs_begin_buffer);
        kernel_bfs_fwd_iteration->setArg("n_succs", *n_succs_buffer);
        kernel_bfs_fwd_iteration->setArg("succs", *succs_buffer);

        kernel_bfs_bwd_iteration->setArg("change", *driver->return_value_buffer);
        kernel_bfs_bwd_iteration->setArg("bits", *arena->bits_buffer);
        kernel_bfs_bwd_iteration->setArg("preds_begin", *preds_begin_buffer);
        kernel_bfs_bwd_iteration->setArg("n_preds", *n_preds_buffer);
        kernel_bfs_bwd_iteration->setArg("preds", *preds_buffer);

        kernel_scc_partition_init->setArg("bits", *arena->bits_buffer);
        kernel_scc_partition_init->setArg("pivots", *pivots_buffer);
        kernel_scc_partition_init->setArg("scc_values", *scc_values_buffer);
        kernel_scc_partition_init->setArg("scc_intervals", *scc_intervals_buffer);

        kernel_scc_partition_update_values->setArg("change", *driver->return_value_buffer);
        kernel_scc_partition_update_values->setArg("bits", *arena->bits_buffer);
        kernel_scc_partition_update_values->setArg("scc_values", *scc_values_buffer);
        kernel_scc_partition_update_values->setArg("scc_intervals", *scc_intervals_buffer);

        kernel_scc_partition_update_edges->setArg("bits", *arena->bits_buffer);
        kernel_scc_partition_update_edges->setArg("succs_begin", *succs_begin_buffer);
        kernel_scc_partition_update_edges->setArg("n_succs", *n_succs_buffer);
        kernel_scc_partition_update_edges->setArg("preds_begin", *preds_begin_buffer);
        kernel_scc_partition_update_edges->setArg("n_preds", *n_preds_buffer);
        kernel_scc_partition_update_edges->setArg("succs", *succs_buffer);
        kernel_scc_partition_update_edges->setArg("preds", *preds_buffer);
        kernel_scc_partition_update_edges->setArg("pivots", *pivots_buffer);
        kernel_scc_partition_update_edges->setArg("scc_values", *scc_values_buffer);

        kernel_trivial_sccs->setArg("change", *driver->return_value_buffer);
        kernel_trivial_sccs->setArg("bits", *arena->bits_buffer);
        kernel_trivial_sccs->setArg("n_succs", *n_succs_buffer);
        kernel_trivial_sccs->setArg("n_preds", *n_preds_buffer);
    }
}

void OCLPreprocessor::clean() {
    if (compute_scc_partition) {
        compute::copy(scc_ids_buffer->cbegin(), scc_ids_buffer->cend(), scc_ids, driver->queue);

        delete n_succs_buffer;
        delete n_preds_buffer;
        delete succs_buffer;
        delete preds_buffer;

        delete pivots_buffer;
        delete scc_ids_buffer;
        delete scc_values_buffer;
        delete scc_intervals_buffer;
    }
}

void OCLPreprocessor::execute_module() {
    if (compute_scc_partition) {
        scc_decomposition();
    }
}

void OCLPreprocessor::scc_decomposition() {
    driver->executeVoidKernel(kernel_scc_partition_init);
    print_values_debug();
    bool change = true;
    while (change) {
        bfs();
        change = driver->executeBoolKernel(kernel_scc_partition_update_values);
        driver->executeVoidKernel(kernel_scc_partition_update_edges);
        print_values_debug();
    }
}

void OCLPreprocessor::bfs() {
    print_debug("Executing bfs…");
    driver->executeVoidKernel(kernel_bfs_init);
    print_values_debug();
    bool change = true;
    while (change) {
        print_debug("Executing bfs fwd iteration…");
        change = driver->executeBoolKernel(kernel_bfs_fwd_iteration);
        print_values_debug();
    }
    change = true;
    while (change) {
        print_debug("Executing bfs bwd iteration…");
        change = driver->executeBoolKernel(kernel_bfs_bwd_iteration);
        print_values_debug();
    }
}

void OCLPreprocessor::trivial_sccs() {
    bool change = true;
    while (change) {
        print_debug("Finding trivial SCCs…");
        change = driver->executeBoolKernel(kernel_trivial_sccs);
    }
}

inline void OCLPreprocessor::print_values_debug() {
#ifndef NDEBUG
    bits_t*       bits          = new bits_t      [arena->n_nodes];
    id_t*         pivots        = new id_t        [hashtable_size];
    id_t*         scc_ids       = new id_t        [arena->n_nodes];
    scc_value*    scc_values    = new scc_value   [arena->n_nodes];
    scc_interval* scc_intervals = new scc_interval[arena->n_nodes];

    compute::copy(arena->bits_buffer->cbegin(), arena->bits_buffer->cend(), bits, driver->queue);
    compute::copy(pivots_buffer->cbegin(), pivots_buffer->cend(), pivots, driver->queue);
    compute::copy(scc_ids_buffer->cbegin(), scc_ids_buffer->cend(), scc_ids, driver->queue);
    compute::copy(scc_values_buffer->cbegin(), scc_values_buffer->cend(), scc_values, driver->queue);
    compute::copy(scc_intervals_buffer->cbegin(), scc_intervals_buffer->cend(), scc_intervals, driver->queue);

    std::cout << "       v  o(v)    scc(v)    p(v)      v(v)>       v(v)      v(v)<" << std::endl;
    for (size_t i = 0; i < arena->n_nodes; ++i) {
        std::cout << std::setw(8) << arena->ids[i];
        if (IS_P0_NODE(bits[i])) {
            std::cout << " P0";
        }
        else {
            std::cout << " P1";
        }
        if (IS_NOT_SOLVED(bits[i])) {
            std::cout << "        ⊥";
        }
        else {
            std::cout << " " << std::setw(8) << arena->ids[scc_ids[i]];
        }
        std::cout << " " << std::setw(8) << pivots[HASH(scc_values[i].value_num, scc_values[i].value_den, hashtable_size)];

        std::stringstream value_str_stream;
        std::stringstream value_str_pre_stream;
        std::stringstream value_str_post_stream;
        std::stringstream value_lower_str_stream;
        std::stringstream value_upper_str_stream;

        value_str_stream << scc_values[i].value_num << "/" << scc_values[i].value_den;

        if (IS_NOT_SOLVED(bits[i])) {
            value_lower_str_stream <<  scc_intervals[i].value_lower_num << "/" << scc_intervals[i].value_lower_den;
            value_upper_str_stream <<  scc_intervals[i].value_upper_num << "/" << scc_intervals[i].value_upper_den;

            if (IS_LOWER_BIT_SET(bits[i])) {
                value_str_pre_stream << "> ";
            }
            else if (
                    scc_values[i].value_num == scc_intervals[i].value_lower_num &&
                    scc_values[i].value_den == scc_intervals[i].value_lower_den
            ) {
                value_str_pre_stream << "= ";
            }

            if (IS_UPPER_BIT_SET(bits[i])) {
                value_str_post_stream << " <";
            }
            else if (
                    scc_values[i].value_num == scc_intervals[i].value_upper_num &&
                    scc_values[i].value_den == scc_intervals[i].value_upper_den
            ) {
                value_str_post_stream << " =";
            }
        }

        std::string value_lower_str = value_lower_str_stream.str();
        value_str_pre_stream << value_str_stream.str();
        std::string value_str = value_str_pre_stream.str();
        std::string value_str_post = value_str_post_stream.str();
        std::string value_upper_str = value_upper_str_stream.str();

        std::cout << " " << std::setw(10) << value_lower_str;
        std::cout << " " << std::setw(10) << value_str << value_str_post;
        std::cout << " " << std::setw(10 - value_str_post.length()) << value_upper_str;
        std::cout << std::endl;
    }

    delete[] bits;
    delete[] pivots;
    delete[] scc_ids;
    delete[] scc_values;
    delete[] scc_intervals;
#endif
}

void OCLPreprocessor::print_partitions(std::ostream& out) const {
    for (size_t i = 0; i < arena->n_nodes; ++i) {
        out << std::setw(8) << arena->ids[i];
        out << " " << std::setw(8) << arena->ids[scc_ids[i]] << std::endl;
    }
}

}
