#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel scc_partition_init(
        global bits_t* bits,
        global id_t* pivots,
        global scc_value* scc_values,
        global scc_interval* scc_intervals
) {
    id_t i = get_global_id(0);
    if (i < N_NODES) {
        RESET_SOLVED(bits[i]);
        scc_values[i].value_num = 1;
        scc_values[i].value_den = 2;
        scc_intervals[i].value_lower_num =  0;
        scc_intervals[i].value_lower_den =  1;
        scc_intervals[i].value_upper_num =  1;
        scc_intervals[i].value_upper_den =  1;
        // TODO better selection of initial pivot?
        pivots[HASH(1, 2, HASHTABLE_SIZE)] = i;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel scc_partition_update_values(
        global bits_t* bits,
        global scc_value* scc_values,
        global scc_interval* scc_intervals,
        global cl_bool* change
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        if (ARE_BOTH_BITS_SET(bits[i])) {
            SET_SOLVED(bits[i]);
        }
        else {
            if (IS_UPPER_BIT_SET(bits[i])) {
                // forward reachable, must be placed after current scc
                scc_intervals[i].value_lower_num = scc_values[i].value_num;
                scc_intervals[i].value_lower_den = scc_values[i].value_den;
            }
            else {
                // not forward reachable, may be placed before current scc
                scc_intervals[i].value_upper_num = scc_values[i].value_num;
                scc_intervals[i].value_upper_den = scc_values[i].value_den;
            }
            scc_values[i].value_num = scc_intervals[i].value_lower_num + scc_intervals[i].value_upper_num;
            scc_values[i].value_den = scc_intervals[i].value_lower_den + scc_intervals[i].value_upper_den;
            *change = true;
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel scc_partition_update_edges(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global id_t* n_succs,
        global const id_t* preds_begin,
        global id_t* n_preds,
        global id_t* succs,
        global id_t* preds,
        global id_t* pivots,
        global const scc_value* scc_values
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        // race to entry in pivot hashtable to select the new pivot element
        pivots[HASH(scc_values[i].value_num, scc_values[i].value_den, HASHTABLE_SIZE)] = i;

        // delete edges between different partitions
        id_t j     = succs_begin[i];
        id_t j_end = j + n_succs[i];
        while (j < j_end) {
            id_t successor = succs[j];
            if (
                    scc_values[successor].value_num != scc_values[i].value_num ||
                    scc_values[successor].value_den != scc_values[i].value_den
            ) {
                j_end--;
                n_succs[i]--;
                if (j_end > j) {
                    succs[j] = succs[j_end];
                }
            }
            else {
                j++;
            }
        }
        j     = preds_begin[i];
        j_end = j + n_preds[i];
        while (j < j_end) {
            id_t predecessor = preds[j];
            if (
                    scc_values[predecessor].value_num != scc_values[i].value_num ||
                    scc_values[predecessor].value_den != scc_values[i].value_den
            ) {
                j_end--;
                n_preds[i]--;
                if (j_end > j) {
                    preds[j] = preds[j_end];
                }
            }
            else {
                j++;
            }
        }
    }
}

// TODO implement
#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel trivial_sccs(
        global const bits_t* bits,
        global const id_t* n_succs,
        global const id_t* n_preds,
        global cl_bool* change
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        //pivots[HASH(scc_values[i].value_num, scc_values[i].value_den, HASHTABLE_SIZE)] = i;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bfs_init(
        global bits_t* bits,
        global const scc_value* scc_values,
        global id_t* scc_ids,
        global const id_t* pivots
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        // Already set id of pivot here so new pivot can be selected during scc partition update
        scc_ids[i] = pivots[HASH(scc_values[i].value_num, scc_values[i].value_den, HASHTABLE_SIZE)];
        if (i == scc_ids[i]) {
            SET_BOTH_BITS(bits[i]);
            RESET_VISITED(bits[i]);
        }
        else {
            RESET_BOTH_BITS(bits[i]);
            SET_VISITED(bits[i]);
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bfs_fwd_iteration(
        global bits_t* bits,
        global const id_t* succs_begin,
        global const id_t* n_succs,
        global const id_t* succs,
        global cl_bool* change
) {
    id_t i = get_global_id(0);
    if (
            i < N_NODES &&
            IS_NOT_SOLVED(bits[i]) &&
            IS_UPPER_BIT_SET(bits[i]) &&
            IS_NOT_FWD_VISITED(bits[i])
    ) {
        for (id_t j = succs_begin[i]; j != succs_begin[i] + n_succs[i]; ++j) {
            id_t successor = succs[j];
            if (IS_UPPER_BIT_NOT_SET(bits[successor])) {
                SET_UPPER_BIT(bits[successor]);
                RESET_FWD_VISITED(bits[successor]);
                *change = true;
            }
        }
        SET_FWD_VISITED(bits[i]);
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bfs_bwd_iteration(
        global bits_t* bits,
        global const id_t* preds_begin,
        global const id_t* n_preds,
        global const id_t* preds,
        global cl_bool* change
) {
    id_t i = get_global_id(0);
    if (
            i < N_NODES &&
            IS_NOT_SOLVED(bits[i]) &&
            IS_LOWER_BIT_SET(bits[i]) &&
            IS_UPPER_BIT_SET(bits[i]) &&
            IS_NOT_BWD_VISITED(bits[i])
    ) {
        for (id_t j = preds_begin[i]; j != preds_begin[i] + n_preds[i]; ++j) {
            id_t predecessor = preds[j];
            if (
                    IS_LOWER_BIT_NOT_SET(bits[predecessor]) &&
                    IS_UPPER_BIT_SET(bits[preds[j]])
            ) {
                SET_LOWER_BIT(bits[predecessor]);
                RESET_BWD_VISITED(bits[predecessor]);
                *change = true;
            }
        }
        SET_BWD_VISITED(bits[i]);
    }
}
