#include "OCLMPGSolver.h"

namespace ocl {

OCLMPGSolver::OCLMPGSolver(OCLDriver* driver, OCLArena* arena, const bool check_for_cycles, const bool zero_mean_partition) :
    OCLSolver(driver, arena),
    check_for_cycles(check_for_cycles),
    zero_mean_partition(zero_mean_partition)
{
    if (!arena->has_edge_weights) {
        throw std::invalid_argument("Arena has no edge weights");
    }
    if (!arena->has_n_succs) {
        arena->computeNumberOfSuccessors();
    }

    driver->addSourceFile("ocl/solver/OCLMPGSolverDefinitions.h");
    driver->addSourceFile("ocl/solver/OCLMPGSolver.cl");

    kernel_ergodic_partition_problem_init = driver->addKernel("ergodic_partition_problem_init");
    kernel_ergodic_partition_problem_update_values = driver->addKernel("ergodic_partition_problem_update_values");
    kernel_ergodic_partition_problem_update_edges = driver->addKernel("ergodic_partition_problem_update_edges");
    kernel_p_mean_partition_problem_init = driver->addKernel("p_mean_partition_problem_init");
    kernel_p_mean_partition_problem_update = driver->addKernel("p_mean_partition_problem_update");
    kernel_bellman_ford_init = driver->addKernel("bellman_ford_init");
    kernel_bellman_ford_iteration = driver->addKernel("bellman_ford_iteration");
    kernel_strategy_improvement = driver->addKernel("strategy_improvement");
    if (check_for_cycles) {
        kernel_cycle_detection_init = driver->addKernel("cycle_detection_init");
        kernel_cycle_detection_iteration = driver->addKernel("cycle_detection_iteration");
        kernel_mark_cycles = driver->addKernel("mark_cycles");
        kernel_propagate_cycles = driver->addKernel("propagate_cycles");
    }

    node_values = new node_value[arena->n_nodes];
}

OCLMPGSolver::~OCLMPGSolver() {
    delete kernel_p_mean_partition_problem_init;
    delete kernel_p_mean_partition_problem_update;
    delete kernel_ergodic_partition_problem_init;
    delete kernel_ergodic_partition_problem_update_values;
    delete kernel_ergodic_partition_problem_update_edges;

    delete kernel_bellman_ford_init;
    delete kernel_bellman_ford_iteration;
    delete kernel_strategy_improvement;

    if (check_for_cycles) {
        delete kernel_cycle_detection_init;
        delete kernel_cycle_detection_iteration;
        delete kernel_mark_cycles;
        delete kernel_propagate_cycles;
    }

    delete[] node_values;
}

void OCLMPGSolver::init() {
    OCLSolver::init();

    succs_begin_buffer    = new compute::vector<id_t>(arena->n_nodes + 1, driver->context);
    n_succs_buffer        = new compute::vector<id_t>(arena->n_nodes, driver->context);
    succs_buffer          = new compute::vector<id_t>(arena->n_edges, driver->context);
    weights_buffer        = new compute::vector<weight_t>(arena->n_edges, driver->context);

    distance_buffer       = new compute::vector<weight_t>(arena->n_nodes, driver->context);
    node_value_buffer     = new compute::vector<node_value>(arena->n_nodes, driver->context);
    successor_alt_buffer  = new compute::vector<id_t>(arena->n_nodes, driver->context);
    node_interval_buffer  = new compute::vector<node_interval>(arena->n_nodes, driver->context);

    compute::copy(arena->succs_begin, arena->succs_begin + arena->n_nodes, succs_begin_buffer->begin(), driver->queue);
    compute::copy(arena->n_succs, arena->n_succs + arena->n_nodes, n_succs_buffer->begin(), driver->queue);
    compute::copy(arena->succs, arena->succs + arena->n_edges, succs_buffer->begin(), driver->queue);
    compute::copy(arena->weights, arena->weights + arena->n_edges, weights_buffer->begin(), driver->queue);

    kernel_ergodic_partition_problem_init->setArg("bits", *arena->bits_buffer);
    kernel_ergodic_partition_problem_init->setArg("node_values", *node_value_buffer);
    kernel_ergodic_partition_problem_init->setArg("node_intervals", *node_interval_buffer);

    kernel_ergodic_partition_problem_update_values->setArg("change", *driver->return_value_buffer);
    kernel_ergodic_partition_problem_update_values->setArg("bits", *arena->bits_buffer);
    kernel_ergodic_partition_problem_update_values->setArg("node_values", *node_value_buffer);
    kernel_ergodic_partition_problem_update_values->setArg("node_intervals", *node_interval_buffer);
    kernel_ergodic_partition_problem_update_values->setArg("succs", *succs_buffer);
    kernel_ergodic_partition_problem_update_values->setArg("successors", *successor_buffer);


    kernel_ergodic_partition_problem_update_edges->setArg("bits", *arena->bits_buffer);
    kernel_ergodic_partition_problem_update_edges->setArg("succs_begin", *succs_begin_buffer);
    kernel_ergodic_partition_problem_update_edges->setArg("n_succs", *n_succs_buffer);
    kernel_ergodic_partition_problem_update_edges->setArg("succs", *succs_buffer);
    kernel_ergodic_partition_problem_update_edges->setArg("weights", *weights_buffer);
    kernel_ergodic_partition_problem_update_edges->setArg("node_values", *node_value_buffer);

    kernel_p_mean_partition_problem_init->setArg("bits", *arena->bits_buffer);
    kernel_p_mean_partition_problem_init->setArg("succs_begin", *succs_begin_buffer);
    kernel_p_mean_partition_problem_init->setArg("n_succs", *n_succs_buffer);
    kernel_p_mean_partition_problem_init->setArg("succs", *succs_buffer);
    kernel_p_mean_partition_problem_init->setArg("distances", *distance_buffer);
    kernel_p_mean_partition_problem_init->setArg("successors", *successor_buffer);

    kernel_p_mean_partition_problem_update->setArg("bits", *arena->bits_buffer);
    kernel_p_mean_partition_problem_update->setArg("succs_begin", *succs_begin_buffer);
    kernel_p_mean_partition_problem_update->setArg("n_succs", *n_succs_buffer);
    kernel_p_mean_partition_problem_update->setArg("succs", *succs_buffer);
    kernel_p_mean_partition_problem_update->setArg("distances", *distance_buffer);
    kernel_p_mean_partition_problem_update->setArg("successors", *successor_buffer);
    kernel_p_mean_partition_problem_update->setArg("successors_alt", *successor_alt_buffer);
    kernel_p_mean_partition_problem_update->setArg("node_values", *node_value_buffer);
    kernel_p_mean_partition_problem_update->setArg("node_intervals", *node_interval_buffer);

    kernel_bellman_ford_init->setArg("bits", *arena->bits_buffer);
    kernel_bellman_ford_init->setArg("distances", *distance_buffer);

    kernel_bellman_ford_iteration->setArg("change", *driver->return_value_buffer);
    kernel_bellman_ford_iteration->setArg("bits", *arena->bits_buffer);
    kernel_bellman_ford_iteration->setArg("succs_begin", *succs_begin_buffer);
    kernel_bellman_ford_iteration->setArg("n_succs", *n_succs_buffer);
    kernel_bellman_ford_iteration->setArg("succs", *succs_buffer);
    kernel_bellman_ford_iteration->setArg("weights", *weights_buffer);
    kernel_bellman_ford_iteration->setArg("distances", *distance_buffer);
    kernel_bellman_ford_iteration->setArg("successors", *successor_buffer);
    kernel_bellman_ford_iteration->setArg("node_values", *node_value_buffer);

    kernel_strategy_improvement->setArg("change", *driver->return_value_buffer);
    kernel_strategy_improvement->setArg("bits", *arena->bits_buffer);
    kernel_strategy_improvement->setArg("succs_begin", *succs_begin_buffer);
    kernel_strategy_improvement->setArg("n_succs", *n_succs_buffer);
    kernel_strategy_improvement->setArg("succs", *succs_buffer);
    kernel_strategy_improvement->setArg("weights", *weights_buffer);
    kernel_strategy_improvement->setArg("distances", *distance_buffer);
    kernel_strategy_improvement->setArg("successors", *successor_buffer);
    kernel_strategy_improvement->setArg("node_values", *node_value_buffer);

    if (check_for_cycles) {
        kernel_cycle_detection_init->setArg("bits", *arena->bits_buffer);
        kernel_cycle_detection_init->setArg("distances", *distance_buffer);
        kernel_cycle_detection_init->setArg("successors", *successor_buffer);
        kernel_cycle_detection_init->setArg("successors_alt", *successor_alt_buffer);

        kernel_cycle_detection_iteration->setArg("change", *driver->return_value_buffer);
        kernel_cycle_detection_iteration->setArg("bits", *arena->bits_buffer);
        kernel_cycle_detection_iteration->setArg("succs_begin", *succs_begin_buffer);
        kernel_cycle_detection_iteration->setArg("n_succs", *n_succs_buffer);
        kernel_cycle_detection_iteration->setArg("succs", *succs_buffer);
        kernel_cycle_detection_iteration->setArg("weights", *weights_buffer);
        kernel_cycle_detection_iteration->setArg("distances", *distance_buffer);
        kernel_cycle_detection_iteration->setArg("successors", *successor_buffer);
        kernel_cycle_detection_iteration->setArg("successors_alt", *successor_alt_buffer);
        kernel_cycle_detection_iteration->setArg("node_values", *node_value_buffer);

        kernel_mark_cycles->setArg("bits", *arena->bits_buffer);
        kernel_mark_cycles->setArg("succs", *succs_buffer);
        kernel_mark_cycles->setArg("weights", *weights_buffer);
        kernel_mark_cycles->setArg("distances", *distance_buffer);
        kernel_mark_cycles->setArg("successors", *successor_buffer);
        kernel_mark_cycles->setArg("successors_alt", *successor_alt_buffer);
        kernel_mark_cycles->setArg("node_values", *node_value_buffer);

        kernel_propagate_cycles->setArg("change", *driver->return_value_buffer);
        kernel_propagate_cycles->setArg("bits", *arena->bits_buffer);
        kernel_propagate_cycles->setArg("succs", *succs_buffer);
        kernel_propagate_cycles->setArg("distances", *distance_buffer);
        kernel_propagate_cycles->setArg("successors", *successor_buffer);
    }
}

void OCLMPGSolver::clean() {
    if (zero_mean_partition) {
        compute::copy(arena->bits_buffer->cbegin(), arena->bits_buffer->cend(), arena->bits, driver->queue);
    }
    else {
        compute::copy(node_value_buffer->cbegin(), node_value_buffer->cend(), node_values, driver->queue);
    }

    delete succs_begin_buffer;
    delete n_succs_buffer;
    delete succs_buffer;
    delete weights_buffer;

    delete distance_buffer;
    delete node_value_buffer;
    delete successor_alt_buffer;
    delete node_interval_buffer;

    OCLSolver::clean();
}

void OCLMPGSolver::execute_module() {
    print_debug("Initializing ergodic partition problem…");
	driver->executeVoidKernel(kernel_ergodic_partition_problem_init);

    if (zero_mean_partition) {
        p_mean_partition_problem();
    }
    else {
        ergodic_partition_problem();
    }
}

void OCLMPGSolver::p_mean_partition_problem(cl_bool invert) {
    print_debug("Initializing p-mean partition problem…");
    kernel_p_mean_partition_problem_init->setArg("invert", invert);
    driver->executeVoidKernel(kernel_p_mean_partition_problem_init);
    print_values_debug(invert);

    strategy_iteration(invert);

    print_debug("Updating partitions");
    kernel_p_mean_partition_problem_update->setArg("invert", invert);
    driver->executeVoidKernel(kernel_p_mean_partition_problem_update);
    print_values_debug(invert, true, false);
}

void OCLMPGSolver::strategy_iteration(cl_bool invert) {
    if (check_for_cycles) {
        kernel_cycle_detection_init->setArg("invert", invert);
        kernel_cycle_detection_iteration->setArg("invert", invert);
        kernel_mark_cycles->setArg("invert", invert);
        kernel_propagate_cycles->setArg("invert", invert);

        cycle_detection();
    }

    kernel_bellman_ford_init->setArg("invert", invert);
    kernel_bellman_ford_iteration->setArg("invert", invert);
    kernel_strategy_improvement->setArg("invert", invert);

    bool change = true;
    while (change) {
        bellman_ford();
        print_values_debug(invert);
        change = strategy_improvement();
        print_values_debug(invert);
    }
}

void OCLMPGSolver::threeway_partition_problem() {
    print_debug("Solving normal three-way partition problem…");
    p_mean_partition_problem(CL_FALSE);

    print_debug("Solving inverted three-way partition problem…");
    p_mean_partition_problem(CL_TRUE);
}

void OCLMPGSolver::ergodic_partition_problem() {
    bool change = true;
    while (change) {
        threeway_partition_problem();
        print_debug("Updating ergodic partition problem…");
        change = driver->executeBoolKernel(kernel_ergodic_partition_problem_update_values);
        driver->executeVoidKernel(kernel_ergodic_partition_problem_update_edges);
        print_values_debug();
    }
}

void OCLMPGSolver::cycle_detection() {
    print_debug("Executing cycle detection algorithm…");
    driver->executeVoidKernel(kernel_cycle_detection_init);
    print_values_debug();
    bool change = true;
    for (size_t i = 0; i < arena->n_nodes && change; ++i) {
        print_debug("Executing cycle detection algorithm iteration…");
        change = driver->executeBoolKernel(kernel_cycle_detection_iteration);
        print_values_debug();
    }
    if (change) {
        print_debug("Executing cycle marking…");
        driver->executeVoidKernel(kernel_mark_cycles);
        print_values_debug();
    }
    while (change) {
        print_debug("Executing cycle propagation…");
        change = driver->executeBoolKernel(kernel_propagate_cycles);
        print_values_debug();
    }
}

void OCLMPGSolver::bellman_ford() {

    print_debug("Executing Bellman-Ford algorithm…");
    driver->executeVoidKernel(kernel_bellman_ford_init);
    bool change = true;
    for (size_t i = 0; i < arena->n_nodes && change; ++i) {
        print_debug("Executing Bellman-Ford iteration…");
        change = driver->executeBoolKernel(kernel_bellman_ford_iteration);
    }
}

bool OCLMPGSolver::strategy_improvement() {
    print_debug("Executing strategy improvement algorithm…");
    return driver->executeBoolKernel(kernel_strategy_improvement);
}

inline void OCLMPGSolver::print_values_debug(cl_bool invert, bool print_min_succ, bool print_max_succ) {
#ifndef NDEBUG
    node_interval* node_intervals = new node_interval[arena->n_nodes];
    weight_t* distances = new weight_t[arena->n_nodes];
    weight_t* weights = new weight_t[arena->n_edges];
    id_t* succs = new id_t[arena->n_edges];

    compute::copy(arena->bits_buffer->cbegin(), arena->bits_buffer->cend(), arena->bits, driver->queue);
    compute::copy(succs_buffer->cbegin(), succs_buffer->cend(), succs, driver->queue);
    compute::copy(weights_buffer->cbegin(), weights_buffer->cend(), weights, driver->queue);
    compute::copy(distance_buffer->cbegin(), distance_buffer->cend(), distances, driver->queue);
    compute::copy(successor_buffer->cbegin(), successor_buffer->cend(), successors, driver->queue);
    compute::copy(node_value_buffer->cbegin(), node_value_buffer->cend(), node_values, driver->queue);
    compute::copy(node_interval_buffer->cbegin(), node_interval_buffer->cend(), node_intervals, driver->queue);

    std::cout << "       v  o(v)     σ(v)       d(v)      v(v)>       v(v)      v(v)<" << std::endl;
    for (size_t i = 0; i < arena->n_nodes; ++i) {
        bool print_succ;
        std::cout << std::setw(8) << arena->ids[i];
        if (IS_P0_NODE(arena->bits[i]) ^ invert) {
            std::cout << "   MAX";
            print_succ = print_max_succ;
        }
        else {
            std::cout << "   MIN";
            print_succ = print_min_succ;
        }
        if (IS_SOLVED(arena->bits[i])) {
            print_succ = true;
        }
        if (print_succ) {
            if (successors[i] == RETREAT_EDGE) {
                std::cout << "        ⊥";
            }
            else {
                id_t successor;
                if (IS_NOT_SOLVED(arena->bits[i])) {
                    successor = arena->succs[successors[i]];
                }
                else {
                    successor = successors[i];
                }
                std::stringstream successor_str_stream;
                if (IS_AUX_EDGE(successor)) {
                    successor_str_stream << "x ";
                    successor &= MASK_NO_AUX_EDGE;
                }
                successor_str_stream << arena->ids[successor];
                std::cout << " " << std::setw(8) << successor_str_stream.str();
            }
        }
        else {
            std::cout << "         ";
        }
        if (distances[i] == DISTANCE_INFINITY) {
            std::cout << "          ∞";
        }
        else if (distances[i] == DISTANCE_MINUS_INFINITY) {
            std::cout << "         -∞";
        }
        else {
            std::cout << " " << std::setw(10) << distances[i];
        }

        std::stringstream value_str_stream;
        std::stringstream value_str_pre_stream;
        std::stringstream value_str_post_stream;
        std::stringstream value_lower_str_stream;
        std::stringstream value_upper_str_stream;

        value_str_stream << node_values[i].value_num << "/" << node_values[i].value_den;

        if (IS_NOT_SOLVED(arena->bits[i])) {
            value_lower_str_stream <<  node_intervals[i].value_lower_num << "/" << node_intervals[i].value_lower_den;
            value_upper_str_stream <<  node_intervals[i].value_upper_num << "/" << node_intervals[i].value_upper_den;

            if (IS_LOWER_BIT_SET(arena->bits[i])) {
                value_str_pre_stream << "> ";
            }
            else if (
                    node_values[i].value_num == node_intervals[i].value_lower_num &&
                    node_values[i].value_den == node_intervals[i].value_lower_den
            ) {
                value_str_pre_stream << "= ";
            }

            if (IS_UPPER_BIT_SET(arena->bits[i])) {
                value_str_post_stream << " <";
            }
            else if (
                    node_values[i].value_num == node_intervals[i].value_upper_num &&
                    node_values[i].value_den == node_intervals[i].value_upper_den
            ) {
                value_str_post_stream << " =";
            }
        }

        std::string value_lower_str = value_lower_str_stream.str();
        value_str_pre_stream << value_str_stream.str();
        std::string value_str = value_str_pre_stream.str();
        std::string value_str_post = value_str_post_stream.str();
        std::string value_upper_str = value_upper_str_stream.str();

        std::cout << " " << std::setw(10) << value_lower_str;
        std::cout << " " << std::setw(10) << value_str << value_str_post;
        std::cout << " " << std::setw(10 - value_str_post.length()) << value_upper_str;
        std::cout << std::endl;
    }

    delete[] succs;
    delete[] distances;
    delete[] node_intervals;
#endif
}

void OCLMPGSolver::print_solution(std::ostream& out) const {
    for (size_t i = 0; i < arena->n_nodes; ++i) {
        out << std::setw(8) << arena->ids[i];
        if (IS_P0_NODE(arena->bits[i])) {
            out << " MAX ";
        }
        else {
            out << " MIN ";
        }
        if (zero_mean_partition) {
            if (IS_LOWER_BIT_SET(arena->bits[i])) {
                out << " " << "≤0";
            }
            else {
                out << " " << ">0";
            }
        }
        else {
            id_t successor = arena->ids[successors[i]];
            out << std::setw(8) << successor;

            std::stringstream value_str_stream;
            value_str_stream << node_values[i].value_num << "/" << node_values[i].value_den;
            out << " " << std::setw(10) << value_str_stream.str();
        }

        out << std::endl;
    }
}

}
