#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel ergodic_partition_problem_init(
        global bits_t* bits,
        global node_value* node_values,
        global node_interval* node_intervals
) {
    id_t i = get_global_id(0);
    if (i < N_NODES) {
        SET_BOTH_BITS(bits[i]);
        RESET_SOLVED(bits[i]);
        node_values[i].value_num = 0;
        node_values[i].value_den = 1;
        node_intervals[i].value_upper_num = +1;
        node_intervals[i].value_upper_den =  0;
        node_intervals[i].value_lower_num = -1;
        node_intervals[i].value_lower_den =  0;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel ergodic_partition_problem_update_values(
        global bits_t* bits,
        global node_value* node_values,
        global node_interval* node_intervals,
        global const id_t* succs,
        global id_t* successors,
        global cl_bool* change
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        if (
                node_intervals[i].value_lower_num == node_intervals[i].value_upper_num &&
                node_intervals[i].value_lower_den == node_intervals[i].value_upper_den
        ) {
            // node solved
            SET_SOLVED(bits[i]);
            // set successor to actual node
            successors[i] = succs[successors[i]] & MASK_NO_AUX_EDGE;
        }
        else {
            node_values[i].value_num = node_intervals[i].value_lower_num + node_intervals[i].value_upper_num;
            node_values[i].value_den = node_intervals[i].value_lower_den + node_intervals[i].value_upper_den;

            if (IS_LOWER_BIT_SET(bits[i])) {
                // accelerate search along left tree branch towards lower value
                node_intervals[i].value_lower_num *= 2;
                node_intervals[i].value_lower_den *= 2;
            }
            else if (IS_UPPER_BIT_SET(bits[i])) {
                // accelerate search along right tree branch towards upper value
                node_intervals[i].value_upper_num *= 2;
                node_intervals[i].value_upper_den *= 2;
            }
            else if (
                    node_intervals[i].value_lower_den * node_intervals[i].value_upper_num -
                    node_intervals[i].value_lower_num * node_intervals[i].value_upper_den == 1
            ) {
                // in between two nodes on the tree branch, descend into subtree
                SET_BOTH_BITS(bits[i]);
            }
            else {
                // binary search in finite interval on the tree branch
                // need to reduce fraction due to value_lower and value_upper not being Farey neighbours
                node_values[i].value_num /= 2;
                node_values[i].value_den /= 2;
            }

            *change = true;
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel ergodic_partition_problem_update_edges(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global id_t* n_succs,
        global const node_value* node_values,
        global id_t* succs,
        global weight_t* weights
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        // delete edges between partitions
        id_t j     = succs_begin[i];
        id_t j_end = j + n_succs[i];
        while (j < j_end) {
            id_t successor = succs[j];
            if (
                    node_values[successor].value_num != node_values[i].value_num ||
                    node_values[successor].value_den != node_values[i].value_den
            ) {
                // nodes in different partitions, delete edge
                j_end--;
                n_succs[i]--;
                if (j_end > j) {
                    succs[j] = succs[j_end];
                    weights[j] = weights[j_end];
                }
            }
            else {
                j++;
            }
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel p_mean_partition_problem_init(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global const id_t* n_succs,
        global weight_t* distances,
        global id_t* successors,
        global id_t* succs,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        successors[i] = RETREAT_EDGE;
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
        if (
            (IS_P1_NODE(bits[i]) ^ invert) &&
#if N_NODES_AUXILLIARY_FWD && !N_NODES_AUXILLIARY_BWD
            IS_FWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && !N_NODES_AUXILLIARY_FWD
            IS_BWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && N_NODES_AUXILLIARY_FWD
            IS_AUX_NODE(bits[i])
#endif
        ) {
            for (id_t j = succs_begin[i]; j != succs_begin[i] + n_succs[i]; ++j) {
                id_t successor = succs[j];
                if (
#if N_NODES_AUXILLIARY_FWD && !N_NODES_AUXILLIARY_BWD
                         (successor >= i) &&
#elif N_NODES_AUXILLIARY_BWD && !N_NODES_AUXILLIARY_FWD
                        (successor <= i) &&
#endif
                        (IS_P1_NODE(bits[successor]) ^ invert)
                ) {
                    SET_AUX_EDGE(succs[j]);
                }
            }
        }
#endif
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel p_mean_partition_problem_update(
        global bits_t* bits,
        global const id_t* succs_begin,
        global const id_t* n_succs,
        global const weight_t* distances,
        global id_t* successors,
        global id_t* successors_alt,
        global const node_value* node_values,
        global node_interval* node_intervals,
        global id_t* succs,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        if ((distances[i] == DISTANCE_INFINITY) ^ invert) {
            node_intervals[i].value_lower_num = node_values[i].value_num;
            node_intervals[i].value_lower_den = node_values[i].value_den;
            RESET_LOWER_BIT(bits[i]);
        }
        else {
            node_intervals[i].value_upper_num = node_values[i].value_num;
            node_intervals[i].value_upper_den = node_values[i].value_den;
            RESET_UPPER_BIT(bits[i]);
        }
        if (IS_P1_NODE(bits[i]) ^ invert) {
            // save successor
            successors_alt[i] = successors[i];
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
            if (
#if N_NODES_AUXILLIARY_FWD && !N_NODES_AUXILLIARY_BWD
                IS_FWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && !N_NODES_AUXILLIARY_FWD
                IS_BWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && N_NODES_AUXILLIARY_FWD
                IS_AUX_NODE(bits[i])
#endif
            ) {
                for (id_t j = succs_begin[i]; j != succs_begin[i] + n_succs[i]; ++j) {
                    RESET_AUX_EDGE(succs[j]);
                }
            }
#endif
        }
        else {
            // restore successor
            successors[i] = successors_alt[i];
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel cycle_detection_init(
        global const bits_t* bits,
        global weight_t* distances,
        global id_t* successors,
        global id_t* successors_alt,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i]) && (IS_P1_NODE(bits[i]) ^ invert)) {
        distances[i] = 0;
        successors_alt[i] = 0;
        successors[i] = RETREAT_EDGE;
    }
}

inline weight_t calculate_succ_distance(
        const id_t node,
        const id_t edge,
        const cl_bool invert,
        global const bits_t* bits,
        global const id_t* succs,
        global const weight_t* weights,
        global const weight_t* distances,
        global const node_value* node_values
) {
    if (edge == RETREAT_EDGE) {
        return 0;
    }
    else {
        id_t successor = succs[edge];
        weight_t succ_distance;
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
        if (IS_AUX_EDGE(successor)) {
            succ_distance = 0;
        }
        else {
#endif
            succ_distance = distances[successor];
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
        }
#endif
        if (
                succ_distance != DISTANCE_INFINITY &&
                succ_distance != DISTANCE_MINUS_INFINITY
        ) {
            weight_t addition = node_values[node].value_den * weights[edge] - node_values[node].value_num;
            if (invert) {
                succ_distance -= addition;
            }
            else {
                succ_distance += addition;
            }
        }
        return succ_distance;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel cycle_detection_iteration(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global const id_t* n_succs,
        global weight_t* distances,
        global id_t* successors,
        global id_t* successors_alt,
        global const node_value* node_values,
        global const id_t* succs,
        global const weight_t* weights,
        global cl_bool* change,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i]) && (IS_P1_NODE(bits[i]) ^ invert)) {
        for (id_t j = succs_begin[i]; j != succs_begin[i] + n_succs[i]; ++j) {
            id_t successor = succs[j];
            if (IS_P1_NODE(bits[successor]) ^ invert) {
                weight_t succ_distance = calculate_succ_distance(i, j, invert, bits, succs, weights, distances, node_values);
                weight_t diff = distances[i] - succ_distance;
                if (diff > 0 || (diff == 0 && successors_alt[i] <= successors_alt[successor])) {
                    if (diff >= 0) {
                        successors_alt[i] = successors_alt[successor] + 1;
                    }
                    else {
                        distances[i] = succ_distance;
                    }
                    successors[i] = j;
                    *change = true;
                }
            }
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel mark_cycles(
        global const bits_t* bits,
        global weight_t* distances,
        global const id_t* successors,
        global const id_t* successors_alt,
        global const node_value* node_values,
        global const id_t* succs,
        global const weight_t* weights,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i]) && (IS_P1_NODE(bits[i]) ^ invert)) {
        id_t j = successors[i];
        if (j != RETREAT_EDGE) {
            id_t successor = succs[j];
            weight_t succ_distance = calculate_succ_distance(i, j, invert, bits, succs, weights, distances, node_values);
            if (
                    distances[i] > succ_distance ||
                    (distances[i] == succ_distance &&
                     successors_alt[i] <= successors_alt[successor])
            ) {
                distances[i] = DISTANCE_MINUS_INFINITY;
            }
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel propagate_cycles(
        global const bits_t* bits,
        global weight_t* distances,
        global const id_t* successors,
        global const id_t* succs,
        global cl_bool* change,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (
            i < N_NODES &&
            IS_NOT_SOLVED(bits[i]) &&
            (IS_P1_NODE(bits[i]) ^ invert) &&
            successors[i] != RETREAT_EDGE &&
            distances[i] != DISTANCE_MINUS_INFINITY &&
            distances[succs[successors[i]]] == DISTANCE_MINUS_INFINITY
    ) {
        distances[i] = DISTANCE_MINUS_INFINITY;
        *change = true;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_init(
        global const bits_t* bits,
        global weight_t* distances,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (
            i < N_NODES &&
            IS_NOT_SOLVED(bits[i]) &&
            (IS_P0_NODE(bits[i]) ^ invert ||
             distances[i] != DISTANCE_MINUS_INFINITY)
    ) {
        distances[i] = DISTANCE_INFINITY;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_iteration(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global const id_t* n_succs,
        global weight_t* distances,
        global id_t* successors,
        global const node_value* node_values,
        global const id_t* succs,
        global const weight_t* weights,
        global cl_bool* change,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        id_t j;
        id_t j_end;
        if (IS_P1_NODE(bits[i]) ^ invert) {
            j     = succs_begin[i];
            j_end = j + n_succs[i];
        }
        else {
            j     = successors[i];
            j_end = j + 1;
        }

        for (; j != j_end; ++j) {
            weight_t succ_distance = calculate_succ_distance(i, j, invert, bits, succs, weights, distances, node_values);
            if (distances[i] > succ_distance) {
                distances[i] = succ_distance;
                successors[i] = j;
                *change = true;
            }
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel strategy_improvement(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global const id_t* n_succs,
        global weight_t* distances,
        global id_t* successors,
        global const node_value* node_values,
        global id_t* succs,
        global const weight_t* weights,
        global cl_bool* change,
        const cl_bool invert
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        if (IS_P0_NODE(bits[i]) ^ invert) {
            for (id_t j = succs_begin[i]; j != succs_begin[i] + n_succs[i]; ++j) {
                weight_t succ_distance = calculate_succ_distance(i, j, invert, bits, succs, weights, distances, node_values);
                if (distances[i] < succ_distance) {
                    distances[i] = succ_distance;
                    successors[i] = j;
                    *change = true;
                }
            }
        }
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
        else if (
#if N_NODES_AUXILLIARY_FWD && !N_NODES_AUXILLIARY_BWD
            IS_FWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && !N_NODES_AUXILLIARY_FWD
            IS_BWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && N_NODES_AUXILLIARY_FWD
            IS_AUX_NODE(bits[i])
#endif
        ) {
            for (id_t j = succs_begin[i]; j != succs_begin[i] + n_succs[i]; ++j) {
                id_t successor = succs[j];
                if (IS_AUX_EDGE(successor)) {
                    weight_t succ_distance = distances[successor & MASK_NO_AUX_EDGE];
                    if (0 < succ_distance) {
                        RESET_AUX_EDGE(succs[j]);
                        *change = true;
                    }
                }
            }
        }
#endif
    }
}
