#include "OCLPGSolver.h"

namespace ocl {

OCLPGSolver::OCLPGSolver(OCLDriver* driver, OCLArena* arena, const bool check_realizability) :
    OCLSolver(driver, arena),
    check_realizability(check_realizability)
{
    if (!arena->has_node_colors && !arena->has_edge_colors) {
        throw std::invalid_argument("Arena has no node or edge colors");
    }
    if (check_realizability && arena->initial_node >= arena->n_nodes) {
        throw std::invalid_argument("No initial node given");
    }

    driver->addSourceFile("ocl/solver/OCLPGSolverDefinitions.h");
    driver->addSourceFile("ocl/solver/OCLPGSolver.cl");

    kernel_strategy_iteration_init = driver->addKernel("strategy_iteration_init");
    kernel_strategy_iteration_update = driver->addKernel("strategy_iteration_update");
    kernel_bellman_ford_update = driver->addKernel("bellman_ford_update");
    kernel_bellman_ford_init = driver->addKernel("bellman_ford_init");
    kernel_bellman_ford_iteration = driver->addKernel("bellman_ford_iteration");
    kernel_strategy_improvement = driver->addKernel("strategy_improvement");
}

OCLPGSolver::~OCLPGSolver() {
    delete kernel_strategy_iteration_init;
    delete kernel_bellman_ford_update;
    delete kernel_bellman_ford_init;
    delete kernel_bellman_ford_iteration;
    delete kernel_strategy_improvement;
}

void OCLPGSolver::init() {
    OCLSolver::init();

    succs_buffer          = new compute::vector<id_t>(arena->n_edges + 1, driver->context);
    succs_begin_buffer    = new compute::vector<id_t>(arena->n_nodes + 1, driver->context);

    compute::copy(arena->succs, arena->succs + arena->n_edges, succs_buffer->begin(), driver->queue);
    compute::copy(arena->succs_begin, arena->succs_begin + arena->n_nodes + 1, succs_begin_buffer->begin(), driver->queue);

    // add retreat edge
    boost::compute::fill_n(succs_buffer->begin() + arena->n_edges, 1, arena->n_nodes, driver->queue);

    size_t n = 0;
    if (arena->has_node_colors) {
        n = arena->n_nodes;
    }
    else if (arena->has_edge_colors) {
        n = arena->n_edges;
    }
    colors_buffer = new compute::vector<color_t>((n + 1) * arena->n_colors, driver->context);
    compute::copy(arena->colors, arena->colors + n, colors_buffer->begin(), driver->queue);

    // set color of retreat edge/node
    boost::compute::fill_n(colors_buffer->begin() + n, 1, arena->neutral_color, driver->queue);

    distance_buffer       = new compute::vector<distance_t>((arena->n_nodes + 1) * arena->n_colors, driver->context);
    distance_prime_buffer = new compute::vector<distance_t>((arena->n_nodes + 1) * arena->n_colors, driver->context);

    // set distance of retreat node to 0
    boost::compute::fill_n(distance_buffer->begin() + (arena->n_nodes * arena->n_colors), arena->n_colors, 0, driver->queue);
    boost::compute::fill_n(distance_prime_buffer->begin() + (arena->n_nodes * arena->n_colors), arena->n_colors, 0, driver->queue);

    kernel_strategy_iteration_init->setArg("bits", *arena->bits_buffer);
    kernel_strategy_iteration_init->setArg("successors", *successor_buffer);
    kernel_strategy_iteration_init->setArg("succs_begin", *succs_begin_buffer);
    kernel_strategy_iteration_init->setArg("succs", *succs_buffer);

    kernel_strategy_iteration_update->setArg("bits", *arena->bits_buffer);
    kernel_strategy_iteration_update->setArg("successors", *successor_buffer);
    kernel_strategy_iteration_update->setArg("succs", *succs_buffer);

    kernel_bellman_ford_update->setArg("bits", *arena->bits_buffer);
    kernel_bellman_ford_update->setArg("distances", *distance_buffer);
    kernel_bellman_ford_update->setArg("succs", *succs_buffer);
    kernel_bellman_ford_update->setArg("successors", *successor_buffer);


    kernel_bellman_ford_init->setArg("bits", *arena->bits_buffer);
    kernel_bellman_ford_init->setArg("distances", *distance_buffer);

    kernel_bellman_ford_iteration->setArg("change", *driver->return_value_buffer);
    kernel_bellman_ford_iteration->setArg("bits", *arena->bits_buffer);
    kernel_bellman_ford_iteration->setArg("succs_begin", *succs_begin_buffer);
    kernel_bellman_ford_iteration->setArg("succs", *succs_buffer);
    kernel_bellman_ford_iteration->setArg("colors", *colors_buffer);
    kernel_bellman_ford_iteration->setArg("distances", *distance_buffer);
    kernel_bellman_ford_iteration->setArg("distances_prime", *distance_prime_buffer);
    kernel_bellman_ford_iteration->setArg("successors", *successor_buffer);

    kernel_strategy_improvement->setArg("change", *driver->return_value_buffer);
    kernel_strategy_improvement->setArg("bits", *arena->bits_buffer);
    kernel_strategy_improvement->setArg("succs_begin", *succs_begin_buffer);
    kernel_strategy_improvement->setArg("succs", *succs_buffer);
    kernel_strategy_improvement->setArg("colors", *colors_buffer);
    kernel_strategy_improvement->setArg("distances", *distance_buffer);
    kernel_strategy_improvement->setArg("successors", *successor_buffer);
}

void OCLPGSolver::clean() {
    compute::copy(arena->bits_buffer->cbegin(), arena->bits_buffer->cend(), arena->bits, driver->queue);

    delete succs_buffer;
    delete succs_begin_buffer;
    delete colors_buffer;
    delete distance_buffer;
    delete distance_prime_buffer;

    OCLSolver::clean();
}

void OCLPGSolver::execute_module() {
    strategy_iteration();
}

void OCLPGSolver::strategy_iteration() {
    print_debug("Initializing strategy iteration…");
	driver->executeVoidKernel(kernel_strategy_iteration_init);

    bool change = true;
    bool prime = false;
    while (change) {
        prime = bellman_ford(prime);
        change = strategy_improvement(prime);
        print_values_debug(prime);
    }

    if (!check_realizability) {
        print_debug("Updating strategy iteration…");
        driver->executeVoidKernel(kernel_strategy_iteration_update);
    }
}

bool OCLPGSolver::bellman_ford(bool prime) {
    print_debug("Executing Bellman-Ford algorithm…");
    if (prime) {
        kernel_bellman_ford_init->setArg("distances", *distance_prime_buffer);
    }
    else {
        kernel_bellman_ford_init->setArg("distances", *distance_buffer);
    }
    driver->executeVoidKernel(kernel_bellman_ford_init);
    print_values_debug(prime);
    bool change = true;
    for (size_t i = 0; i < arena->n_nodes && change; ++i) {
        if (prime) {
            kernel_bellman_ford_iteration->setArg("distances", *distance_prime_buffer);
            kernel_bellman_ford_iteration->setArg("distances_prime", *distance_buffer);
        }
        else {
            kernel_bellman_ford_iteration->setArg("distances", *distance_buffer);
            kernel_bellman_ford_iteration->setArg("distances_prime", *distance_prime_buffer);
        }
        print_debug("Executing Bellman-Ford iteration…");
        change = driver->executeBoolKernel(kernel_bellman_ford_iteration);
        prime = !prime;
        print_values_debug(prime);
    }
    print_debug("Updating distance valuation…");
    if (prime) {
        kernel_bellman_ford_update->setArg("distances", *distance_prime_buffer);
    }
    else {
        kernel_bellman_ford_update->setArg("distances", *distance_buffer);
    }
	driver->executeVoidKernel(kernel_bellman_ford_update);
    print_values_debug(prime);
    return prime;
}

bool OCLPGSolver::strategy_improvement(bool prime) {
    print_debug("Executing strategy improvement algorithm…");
    if (prime) {
        kernel_strategy_improvement->setArg("distances", *distance_prime_buffer);
    }
    else {
        kernel_strategy_improvement->setArg("distances", *distance_buffer);
    }
    return driver->executeBoolKernel(kernel_strategy_improvement);
}

inline void OCLPGSolver::print_values_debug(bool prime) {
#ifndef NDEBUG
    std::cout << "Prime: " << prime << std::endl;
    id_t* succs_begin = new id_t[arena->n_nodes];
    id_t* succs = new id_t[arena->n_edges];
    distance_t* distances = new distance_t[arena->n_nodes * arena->n_colors];

    compute::copy(arena->bits_buffer->cbegin(), arena->bits_buffer->cend(), arena->bits, driver->queue);
    if (prime) {
        compute::copy(distance_prime_buffer->cbegin(), distance_prime_buffer->cbegin() + (arena->n_nodes * arena->n_colors), distances, driver->queue);
        compute::copy(distance_prime_buffer->cbegin(), distance_prime_buffer->cbegin() + (arena->n_nodes * arena->n_colors), distances, driver->queue);
    }
    else {
        compute::copy(distance_buffer->cbegin(), distance_buffer->cbegin() + (arena->n_nodes * arena->n_colors), distances, driver->queue);
    }
    compute::copy(succs_begin_buffer->cbegin(), succs_begin_buffer->cbegin() + arena->n_nodes, succs_begin, driver->queue);
    compute::copy(succs_buffer->cbegin(), succs_buffer->cbegin() + arena->n_edges, succs, driver->queue);
    compute::copy(successor_buffer->cbegin(), successor_buffer->cbegin() + arena->n_nodes, successors, driver->queue);

    std::cout << "        v  o(v)     σ(v)";
    for (color_t c = 0; c < arena->n_colors; ++c) {
        std::cout << " d(v)[" << std::setw(4) << c << "]";
    }
    std::cout << std::endl;
    for (size_t i = 0; i < arena->n_nodes; ++i) {
        if (IS_SOLVED(arena->bits[i])) {
            std::cout << "x";
        }
        else {
            std::cout << " ";
        }
        std::cout << std::setw(8) << arena->ids[i];
        if (IS_P0_NODE(arena->bits[i])) {
            std::cout << "   EVE";
        }
        else {
            std::cout << "   ODD";
        }
        if (successors[i] == arena->n_edges) {
            std::cout << "        ⊥";
        }
        else {
            id_t successor;
            if (IS_NOT_SOLVED(arena->bits[i])) {
                successor = arena->succs[successors[i]];
            }
            else {
                successor = successors[i];
            }
            std::stringstream successor_str_stream;
            if (IS_AUX_EDGE(successor)) {
                successor_str_stream << "x ";
                successor &= MASK_NO_AUX_EDGE;
            }
            successor_str_stream << arena->ids[successor];
            std::cout << " " << std::setw(8) << successor_str_stream.str();
        }
        for (color_t c = 0; c < arena->n_colors; c++) {
            distance_t distance = distances[i*arena->n_colors + c];
            if (distance == COUNT_INFINITY) {
                std::cout << "          ∞";
            }
            else if (distance == COUNT_MINUS_INFINITY) {
                std::cout << "         -∞";
            }
            else {
                std::cout << std::setw(11) << distance;
            }
        }
        std::cout << std::endl;
    }

    delete[] succs_begin;
    delete[] succs;
    delete[] distances;
#endif
}

void OCLPGSolver::print_solution(std::ostream& out) const {
    if (check_realizability) {
        if (IS_SOLVED(arena->bits[arena->initial_node])) {
            out << "REALIZABLE" << std::endl;
            for (size_t i = 0; i < arena->n_nodes; ++i) {
                if (IS_P1_NODE(arena->bits[i]) && IS_SOLVED(arena->bits[i])) {
                    out << arena->ids[i] << " " << arena->ids[successors[i]] << std::endl;
                }
            }
        }
        else {
            out << "UNREALIZABLE" << std::endl;
        }
    }
    else {
        out << "paritysol " << (arena->n_nodes-1) << ";" << std::endl;
        for (size_t i = 0; i < arena->n_nodes; ++i) {
            out << arena->ids[i];
            int owner;
            int parity;
            if (IS_P0_NODE(arena->bits[i])) {
                owner = 0;
            }
            else {
                owner = 1;
            }
            if (IS_SOLVED(arena->bits[i])) {
                parity = 1;
            }
            else {
                parity = 0;
            }
            out << " " << parity;
            if (owner == parity) {
                out << " " << arena->ids[successors[i]];
            }
            out << ";" << std::endl;
        }
    }
}

}
