#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel strategy_iteration_init(
        global bits_t* bits,
        global id_t* successors,
        global const id_t* succs_begin,
        global id_t* succs
) {
    id_t i = get_global_id(0);
    if (i < N_NODES) {
        RESET_SOLVED(bits[i]);
        successors[i] = N_EDGES;
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
        if (
            IS_P0_NODE(bits[i]) &&
#if N_NODES_AUXILLIARY_FWD && !N_NODES_AUXILLIARY_BWD
            IS_FWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && !N_NODES_AUXILLIARY_FWD
            IS_BWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && N_NODES_AUXILLIARY_FWD
            IS_AUX_NODE(bits[i])
#endif
        ) {
            for (id_t j = succs_begin[i]; j != succs_begin[i+1]; j++) {
                id_t successor = succs[j];
                if (
#if N_NODES_AUXILLIARY_FWD && !N_NODES_AUXILLIARY_BWD
                         (successor >= i) &&
#elif N_NODES_AUXILLIARY_BWD && !N_NODES_AUXILLIARY_FWD
                        (successor <= i) &&
#endif
                        (IS_P0_NODE(bits[successor]))
                ) {
                    SET_AUX_EDGE(succs[j]);
                }
            }
        }
#endif
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_update(
        global bits_t* bits,
        global distance_t* distances,
        global const id_t* succs,
        global id_t* successors
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
#ifdef MAX_PARITY
        if (distances[i * N_COLORS + N_COLORS - 1] == COUNT_INFINITY) {
#else //MIN_PARITY
        if (distances[i * N_COLORS] == COUNT_INFINITY) {
#endif
            // node won py p1
            SET_SOLVED(bits[i]);
            // set successor to actual node
            successors[i] = succs[successors[i]] & MASK_NO_AUX_EDGE;
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel strategy_iteration_update(
        global const bits_t* bits,
        global const id_t* succs,
        global id_t* successors
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        // node not marked solved before, therefore won by p0
        // set successor to actual node
        successors[i] = succs[successors[i]] & MASK_NO_AUX_EDGE;
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_init(
        global const bits_t* bits,
        global distance_t* distances
) {
    id_t i = get_global_id(0);
    if (i < N_NODES) {
#ifdef MAX_PARITY
        distances[(i+1) * N_COLORS - 1] = COUNT_INFINITY;
#else //MIN_PARITY
        distances[i * N_COLORS] = COUNT_INFINITY;
#endif
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel bellman_ford_iteration(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global const id_t* succs,
        global const color_t* colors,
        global const distance_t* distances,
        global distance_t* distances_prime,
        global id_t* successors,
        global cl_bool* change
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        id_t j;
        id_t j_end;
        if (IS_P0_NODE(bits[i])) {
            j     = succs_begin[i];
            j_end = succs_begin[i+1];
        }
        else {
            j     = successors[i];
            j_end = j + 1;
        }

        color_t k;
        color_t cur_color;
        distance_t cur_color_change;
#ifdef NODE_COLORS
        cur_color = colors[i];
        cur_color_change = COLOR_DISTANCE_DELTA(cur_color);
#endif
        distance_t distance[N_COLORS];
        k = i * N_COLORS;
        for (color_t l = 0; l < N_COLORS; l++) {
            distance[l] = distances[k];
            k++;
        }
        while (j != j_end) {
            id_t successor = succs[j];
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
            if (IS_AUX_EDGE(succs[j])) {
                successor = N_NODES;
            }
#endif
#ifdef MAX_PARITY
            k = successor * N_COLORS + N_COLORS - 1;
#else //MIN_PARITY
            k = successor * N_COLORS;
#endif
            if (distances[k] != COUNT_INFINITY) {
                cl_bool local_change = false;
#ifdef EDGE_COLORS
                cur_color = colors[j];
                cur_color_change = COLOR_DISTANCE_DELTA(cur_color);
#endif
                distance[cur_color] -= cur_color_change;
#ifdef MAX_PARITY
                for (color_t l = N_COLORS - 1; l != -1; l--) {
#else //MIN_PARITY
                for (color_t l = 0; l < N_COLORS; l++) {
#endif
                    distance_t d = distance[l];
                    distance_t d_succ = distances[k];
                    if (local_change || d_succ < d) {
                        distance[l] = d_succ;
                        local_change = true;
                    }
                    else if (d_succ != d) {
                        break;
                    }
#ifdef MAX_PARITY
                    k--;
#else //MIN_PARITY
                    k++;
#endif
                }
                distance[cur_color] += cur_color_change;
                if (local_change) {
                    successors[i] = j;
                    *change = true;
                }
            }
            j++;
        }
        k = i * N_COLORS;
        for (color_t l = 0; l < N_COLORS; l++) {
            distances_prime[k] = distance[l];
            k++;
        }
    }
}

#if WORK_GROUP_SIZE
__attribute__((reqd_work_group_size(WORK_GROUP_SIZE, 1, 1)))
#endif
void kernel strategy_improvement(
        global const bits_t* bits,
        global const id_t* succs_begin,
        global id_t* succs,
        global const color_t* colors,
        global const distance_t* distances,
        global id_t* successors,
        global cl_bool* change
) {
    id_t i = get_global_id(0);
    if (i < N_NODES && IS_NOT_SOLVED(bits[i])) {
        if (IS_P1_NODE(bits[i])) {
            color_t k;
            k = i * N_COLORS;
            distance_t distance[N_COLORS];
            for (color_t l = 0; l < N_COLORS; l++) {
                distance[l] = distances[k];
                k++;
            }
            color_t cur_color;
            distance_t cur_color_change;
#ifdef NODE_COLORS
            cur_color = colors[i];
            cur_color_change = COLOR_DISTANCE_DELTA(cur_color);
            distance[cur_color] -= cur_color_change;
#endif
            for (id_t j = succs_begin[i]; j != succs_begin[i+1]; j++) {
                id_t successor = succs[j];
#ifdef EDGE_COLORS
                cur_color = colors[j];
                cur_color_change = COLOR_DISTANCE_DELTA(cur_color);
                distance[cur_color] -= cur_color_change;
#endif
#ifdef MAX_PARITY
                k = successor * N_COLORS + N_COLORS - 1;
#else //MIN_PARITY
                k = successor * N_COLORS;
#endif
#ifdef MAX_PARITY
                for (color_t l = N_COLORS - 1; l != -1; l--) {
#else //MIN_PARITY
                for (color_t l = 0; l < N_COLORS; l++) {
#endif
                    distance_t d = distance[l];
                    distance_t d_succ = distances[k];
                    if (d_succ > d) {
                        // TODO implement maximal strategy
                        successors[i] = j;
                        *change = true;
                        return;
                    }
                    else if (d_succ != d) {
                        break;
                    }
#ifdef MAX_PARITY
                    k--;
#else //MIN_PARITY
                    k++;
#endif
                }
#ifdef EDGE_COLORS
                distance[cur_color] += cur_color_change;
#endif
            }
        }
#if N_NODES_AUXILLIARY_FWD || N_NODES_AUXILLIARY_BWD
        else if (
#if N_NODES_AUXILLIARY_FWD && !N_NODES_AUXILLIARY_BWD
            IS_FWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && !N_NODES_AUXILLIARY_FWD
            IS_BWD_AUX_NODE(bits[i])
#elif N_NODES_AUXILLIARY_BWD && N_NODES_AUXILLIARY_FWD
            IS_AUX_NODE(bits[i])
#endif
        ) {
            for (id_t j = succs_begin[i]; j != succs_begin[i+1]; j++) {
                id_t successor = succs[j];
                if (IS_AUX_EDGE(successor)) {
                    successor &= MASK_NO_AUX_EDGE;
                    color_t k;
#ifdef MAX_PARITY
                    k = successor * N_COLORS + N_COLORS - 1;
#else //MIN_PARITY
                    k = successor * N_COLORS;
#endif
#ifdef MAX_PARITY
                    for (color_t l = N_COLORS - 1; l != -1; l--) {
#else //MIN_PARITY
                    for (color_t l = 0; l < N_COLORS; l++) {
#endif
                        distance_t d_succ = distances[k];
                        if (d_succ > 0) {
                            RESET_AUX_EDGE(succs[j]);
                            *change = true;
                            break;
                        }
                        else if (d_succ != 0) {
                            break;
                        }
#ifdef MAX_PARITY
                        k--;
#else //MIN_PARITY
                        k++;
#endif
                    }
                }
            }
        }
#endif
    }
}
