#pragma once

#include <cstddef>
#include <vector>
#include <tuple>
#include <limits>
#include <unordered_map>

#include "ParserUtil.h"

namespace parser {

typedef unsigned int id_t;
typedef unsigned int player_t;
typedef unsigned int color_t;
typedef long long weight_t;

const static color_t COLOR_UNDEFINED = std::numeric_limits<color_t>::max();
const static weight_t WEIGHT_UNDEFINED = std::numeric_limits<weight_t>::max();
const static player_t PLAYER_0 = 0;
const static player_t PLAYER_1 = 1;

enum class ParityType { MIN, MAX };

class Arena {
public:
    struct edge {
        id_t dst;
        weight_t weight;
        color_t color;
    };
    struct node {
        id_t id;
        color_t color;
        player_t owner;

        std::vector<edge> successors;
    };

    std::vector<node> nodes;
    size_t n_nodes;
    size_t n_edges;

    std::unordered_map<size_t, size_t> id_map;

    const bool has_edge_weights;
    const bool has_node_colors;
    const bool has_edge_colors;
    const ParityType parity_type;
    const bool has_undefined_colors;

    void addNode(const node& node) {
        size_t n_succs = node.successors.size();
        id_map[node.id] = n_nodes;
        n_nodes += 1;
        n_edges += n_succs;
        nodes.push_back(node);
    }

    Arena(bool has_edge_weights, bool has_node_colors, bool has_edge_colors, ParityType parity_type = ParityType::MIN, bool has_undefined_colors = false) :
        n_nodes(0),
        n_edges(0),
        has_edge_weights(has_edge_weights),
        has_node_colors(has_node_colors),
        has_edge_colors(has_edge_colors),
        parity_type(parity_type),
        has_undefined_colors(has_undefined_colors)
    { }

    ~Arena() {}
};

}

BOOST_FUSION_ADAPT_STRUCT(
    parser::Arena::edge,
    (parser::id_t, dst)
    (parser::weight_t, weight)
)

BOOST_FUSION_ADAPT_STRUCT(
    parser::Arena::node,
    (parser::id_t, id)
    (parser::color_t, color)
    (parser::player_t, owner)
    (std::vector<parser::Arena::edge>, successors)
)

