#include "BINParser.h"

#include <iostream>
#include <fstream>
#include <limits>

#include "cmake-portable-endian/endianness_config.h"

namespace parser {

BINParser::BINParser() { }

BINParser::~BINParser() { }

BINParseResult* BINParser::parse(const std::string& file_nodes, const std::string& file_edges) const {
    BINParseResult* result = new BINParseResult();

    uint32_t* id_map;
    node_entry* nodes;
    edge_entry* edges;

    // read nodes

    std::ifstream in_nodes(file_nodes, std::ifstream::binary);
    if (!in_nodes.is_open()) {
        throw std::invalid_argument("Could not open node input file: " + file_nodes);
    }

    uint32_t n_nodes;
    uint32_t initial_node;
    uint32_t first_player;

    if (!in_nodes.read((char *)&n_nodes, sizeof(n_nodes))) {
        throw parser_error("Number of nodes could not be read");
    }
    n_nodes = betoh32(n_nodes);
    if (!in_nodes.read((char *)&initial_node, sizeof(initial_node))) {
        throw parser_error("Initial state could not be read");
    }
    initial_node = betoh32(initial_node);
    if (!in_nodes.read((char *)&first_player, sizeof(first_player))) {
        throw parser_error("Initial state could not be read");
    }
    first_player = betoh32(first_player);

    nodes = new node_entry[n_nodes];
    id_map = new uint32_t[2*n_nodes];

    result->n_nodes = n_nodes;
    result->ids = new uint32_t[n_nodes];
    result->bits = new uint32_t[n_nodes];
    result->succs_begin = new uint32_t[n_nodes + 1];

    if (!in_nodes.read((char *)nodes, n_nodes * sizeof(node_entry))) {
        throw parser_error("Nodes could not be read");
    }
    for (size_t i = 0; i < n_nodes; i++) {
        int32_t node_id = betoh32(nodes[i].node_id);
        uint32_t succs_begin = betoh32(nodes[i].succs_begin);
        result->ids[i] = std::abs(node_id);
        result->succs_begin[i] = succs_begin;
        if (node_id >= 0) {
            node_id = 2*node_id;
            result->ids[i] = node_id;
            result->bits[i] = 1 - first_player;
            id_map[node_id] = i;
        }
        else {
            node_id = -2*node_id + 1;
            result->ids[i] = node_id;
            result->bits[i] = first_player;
            id_map[node_id] = i;
        }
    }
    if (initial_node >= 0) {
        result->initial_node = id_map[2*initial_node];
    }
    else {
        result->initial_node = id_map[-2*initial_node + 1];
    }

    in_nodes.close();

    // read edges

    std::ifstream in_edges(file_edges, std::ifstream::binary);
    if (!in_edges.is_open()) {
        throw std::invalid_argument("Could not open edge input file: " + file_edges);
    }

    uint32_t n_edges;
    uint32_t highest_color;

    if (!in_edges.read((char *)&n_edges, sizeof(n_edges))) {
        throw parser_error("Number of edges could not be read");
    }
    n_edges = betoh32(n_edges);
    if (!in_edges.read((char *)&highest_color, sizeof(highest_color))) {
        throw parser_error("Number of colors could not be read");
    }
    highest_color = betoh32(highest_color);

    result->n_edges = n_edges;
    result->n_colors = highest_color + 1;
    edges = new edge_entry[n_edges];
    result->succs_begin[n_nodes] = n_edges;
    result->succs = new uint32_t[n_edges];
    result->colors = new uint32_t[n_edges];

    if (!in_edges.read((char *)edges, n_edges * sizeof(edge_entry))) {
        throw parser_error("Edges could not be read");
    }
    for (size_t j = 0; j < n_edges; j++) {
        int32_t dst_id = betoh32(edges[j].successor);
        uint32_t color = betoh32(edges[j].color);
        id_t successor;
        if (dst_id >= 0) {
            successor = id_map[2*dst_id];
        }
        else {
            successor = id_map[-2*dst_id+1];
        }

        result->succs[j]  = successor;
        result->colors[j] = color;
    }

    in_edges.close();

    delete[] id_map;
    delete[] nodes;
    delete[] edges;

    return result;
}

}
