#include "MPGArenaParser.h"

#include <iostream>
#include <fstream>
#include <limits>

#include <boost/fusion/adapted/adt/adapt_adt.hpp>
#include <boost/fusion/include/adapt_adt.hpp>
#include <boost/phoenix/operator.hpp>
#include <boost/phoenix/object/construct.hpp>
#include <boost/phoenix/statement/throw.hpp>
#include <boost/algorithm/string.hpp>

namespace parser {

template <typename Iterator>
MPGArenaParser::node_parser<Iterator>::node_parser() : node_parser::base_type(node_rule, "node") {
    quoted_string %= '"' > *~qi::char_('"') > '"';
    id_rule %= qi::uint_;
    owner_rule %= qi::lit("MAX")[qi::_val=PLAYER_0] | qi::lit("MIN")[qi::_val=PLAYER_1];
    weight_rule %= qi::long_long;
    color_rule %= qi::eps[qi::_val=COLOR_UNDEFINED];
    edge_rule %= id_rule > ':' > weight_rule;
    node_rule %= qi::eps > id_rule > color_rule > owner_rule > edge_rule % ',' > qi::omit[-quoted_string] > ';';

    id_rule.name("id");
    owner_rule.name("owner");
    weight_rule.name("weight");
    edge_rule.name("edge");
    node_rule.name("node");

    qi::on_error<qi::fail>
    (
        node_rule
      , phx::throw_(phx::construct<parser_error> (make_error_message(qi::_4, qi::_3, qi::_2)))
    );
}

template <typename Iterator>
MPGArenaParser::comment_parser<Iterator>::comment_parser() : comment_parser::base_type(comment_rule, "comment") {
    comment_rule = '#' > *qi::char_;

    comment_rule.name("comment");

    qi::on_error<qi::fail>(comment_rule, qi::_1);
}

MPGArenaParser::MPGArenaParser() :
    parser(),
    comm_parser()
{ }

MPGArenaParser::~MPGArenaParser() { }

bool MPGArenaParser::parse_line(unsigned int line_number, std::string& line, Arena::node& node) const {
    boost::algorithm::trim_left(line);
    iterator_type begin = line.begin();
    iterator_type end   = line.end();
    qi::phrase_parse(begin, end, comm_parser, qi::space);
    if (begin == end) {
        return false;
    }
    else {
        qi::phrase_parse(begin, end, parser, qi::space, node);
        qi::phrase_parse(begin, end, comm_parser, qi::space);
        if (begin != end) {
            throw parser_error("Trailing unparsed: " + std::string(begin, end));
        }
        return true;
    }
}

Arena* MPGArenaParser::parse(const std::string& filename) const {
    std::ifstream in(filename);
    if (!in.is_open()) {
        throw std::invalid_argument("Could not open input file: " + filename);
    }
    in.unsetf(std::ios::skipws);

    std::string line;

    // read nodes and add them to the arena
    unsigned int line_number = 0;
    Arena* arena = new Arena(true, false, false);

    while (std::getline(in, line)) {
        line_number += 1;
        Arena::node node;
        if (parse_line(line_number, line, node)) {
            arena->addNode(node);
        }
    }
    in.close();

    return arena;
}

}
