#include "PGArena.h"

#include <iostream>
#include <iterator>
#include <map>

namespace pg {

Node::Node(id_t id, Player owner, parity_t parity, boost::optional< std::string > name):
    id(id),
    owner(owner),
    parity(parity),
    name(name)
{ }

Node::~Node() {
    remove_edges();
    unlink();
}

void Node::unlink() {
    NodeHook::unlink();
}

unsigned int Node::remove_edges() {
    unsigned int count = 0;
    succ_edges.clear_and_dispose( [&count] (auto x) { count++; delete x; } );
    pred_edges.clear_and_dispose( [&count] (auto x) { count++; delete x; } );
    return count;
}

Edge::Edge(Node* src, Node* dst) :
    src(src),
    dst(dst)
{
    src->succ_edges.push_back(*this);
    dst->pred_edges.push_back(*this);
}
Edge::~Edge() {
    unlink();
}

void Edge::unlink() {
    EdgeSrcHook::unlink();
    EdgeDstHook::unlink();
}


PGArena::PGArena() :
    nodes(),
    n_nodes(0),
    n_edges(0),
    max_parity(0)
{
}

PGArena::~PGArena() {
    nodes.clear_and_dispose( [] (auto x) { delete x; } );
}

Node* PGArena::add_node(id_t id, Player owner, parity_t parity, boost::optional< std::string > name) {
    Node* node = new Node(id, owner, parity, name);
    nodes.push_back(*node);
    n_nodes++;
    if (parity > max_parity) {
        max_parity = parity;
    }
    return node;
}

Edge* PGArena::add_edge(Node* src, Node* dst) {
    Edge* edge = new Edge(src, dst);
    n_edges++;
    return edge;
}

void PGArena::remove_node(Node* node) {
    n_edges -= node->remove_edges();
    delete node;
    n_nodes--;
}

void PGArena::remove_edge(Edge* edge) {
    delete edge;
    n_edges--;
}


void PGArena::print() const {
    for (const auto &n : nodes) {
        std::cout << n.id << " ";
        if (n.owner == Player::ODD) {
            std::cout << " ODD";
        }
        else {
            std::cout << "EVEN";
        }
        std::cout << " " << n.parity;
        if (n.name) {
            std::cout << " \"" << *n.name << "\"";
        }

        std::cout << " [";
        auto pred_it = n.pred_edges.begin();
        auto pred_end = n.pred_edges.end();
        while (pred_it != pred_end) {
            std::cout << pred_it->src->id;
            ++pred_it;
            if (pred_it != pred_end) {
                std::cout << ",";
            }
        }
        std::cout << "] [";
        auto succ_it = n.succ_edges.begin();
        auto succ_end = n.succ_edges.end();
        while (succ_it != succ_end) {
            std::cout << succ_it->dst->id;
            ++succ_it;
            if (succ_it != succ_end) {
                std::cout << ",";
            }
        }
        std::cout << "]" << std::endl;
    }
}

}


