#include "PGArenaParser.h"

#include <iostream>
#include <fstream>
#include <map>

#include <boost/optional.hpp>
#include <boost/phoenix/operator.hpp>
#include <boost/phoenix/object/construct.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/predicate.hpp>

BOOST_FUSION_ADAPT_STRUCT(
    pg::PGArenaParser::edge,
    (pg::id_t, dst)
)

BOOST_FUSION_ADAPT_STRUCT(
    pg::PGArenaParser::node,
    (pg::id_t, id)
    (pg::parity_t, parity)
    (pg::Player, owner)
    (std::vector<pg::PGArenaParser::edge>, successors)
    (boost::optional< std::string >, name)
)

namespace pg {

template <typename Iterator>
PGArenaParser::node_parser<Iterator>::node_parser() : node_parser::base_type(node_rule, "node") {
    quoted_string %= '"' > *~qi::char_('"') > '"';
    id_rule %= qi::uint_;
    owner_rule %= qi::lit("0")[qi::_val=Player::EVEN] | qi::lit("1")[qi::_val=Player::ODD];
    parity_rule %= qi::ulong_long;
    edge_rule %= id_rule;
    node_rule %= qi::eps > id_rule > parity_rule > owner_rule > edge_rule % ',' > (-quoted_string) > ';';

    id_rule.name("id");
    owner_rule.name("owner");
    parity_rule.name("parity");
    edge_rule.name("edge");
    node_rule.name("node");

    qi::on_error<qi::fail>
    (
        node_rule
      , std::cerr
            << phx::val("Error: Expecting ")
            << qi::_4                                        // what failed?
            << phx::val(" here: \"")
            << phx::construct<std::string>(qi::_3, qi::_2)   // iterators to error-pos, end
            << phx::val("\"")
            << std::endl
    );
}

template <typename Iterator>
PGArenaParser::comment_parser<Iterator>::comment_parser() : comment_parser::base_type(comment_rule, "comment") {
    comment_rule = '#' > *qi::char_;

    comment_rule.name("comment");

    qi::on_error<qi::fail>(comment_rule, qi::_1);
}

PGArenaParser::PGArenaParser() :
    parser(),
    comm_parser()
{ }

PGArenaParser::~PGArenaParser() { }

int PGArenaParser::parse_line(unsigned int line_number, std::string& line, node& node) const {
    boost::algorithm::trim_left(line);
    iterator_type begin = line.begin();
    iterator_type end   = line.end();
    qi::phrase_parse(begin, end, comm_parser, qi::space);
    if (begin == end) {
        return 0;
    }
    else {
        bool r = qi::phrase_parse(begin, end, parser, qi::space, node);
        qi::phrase_parse(begin, end, comm_parser, qi::space);
        if (r && begin != end) {
            std::cerr << "Error: Trailing unparsed: " << std::string(begin, end) << std::endl;
            r = false;
        }
        if (!r) {
            std::cerr << "Error: Parsing failed on line " << line_number << ": " << line << std::endl;
            return -1;
        }
        return 1;
    }
}

boost::optional<PGArena*> PGArenaParser::parse(const std::string& filename) const {
    std::ifstream in(filename);
    if (!in.is_open()) {
        std::cerr << "Error: Could not open input file: " << filename << std::endl;
        return boost::none;
    }
    in.unsetf(std::ios::skipws);

    std::string line;
    std::vector<node> nodes{};

    // read nodes
    unsigned int line_number = 0;
    while (std::getline(in, line)) {
        line_number += 1;
        //ignore lines with parity or start
        if (!(
            boost::starts_with(line, "parity") ||
            boost::starts_with(line, "start") ||
            boost::starts_with(line, "timeout")
        )) {
            node node;
            int r = parse_line(line_number, line, node);
            if (r < 0) {
                return boost::none;
            }
            else if (r > 0) {
                nodes.push_back(node);
            }
        }
    }
    in.close();

    PGArena* arena = new PGArena();
    std::map< unsigned int, Node* > node_map;
    for (auto node : nodes) {
        Node* arena_node = arena->add_node(node.id, node.owner, node.parity, node.name);
        node_map[node.id]  = arena_node;
    }
    for (auto node : nodes) {
        Node* src = node_map[node.id];
        for (auto edge : node.successors) {
            Node* dst = node_map[edge.dst];
            arena->add_edge(src, dst);
        }
    }
    return arena;
}

}
